# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import logging
import sys

import modes


# Eventually import other future tests, like
# pin, performance


logger = logging.getLogger(__name__)


def run(args, module_dir, tmp_dir):
    global logger

    try:
        current_mode = modes.all_modes[args.mode](args, module_dir, tmp_dir)
    except KeyError:
        logger.critical("Unknown run mode `%s`. Choose one of: %s" % (args.mode, ", ".join(modes.all_mode_names)))
        sys.exit(5)

    logger.debug("Running mode .setup()")
    current_mode.setup()
    logger.debug("Running mode .run()")
    current_mode.run()
    logger.debug("Running mode .teardown()")
    current_mode.teardown()
    logger.debug("Mode finished")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import cStringIO
import datetime
import glob
import bz2
import hashfs
import json
import logging
import os
import shutil

logger = logging.getLogger(__name__)


class CertDB(object):
    """
    Class to efficiently store SSL certificates
    """
    # TODO: CertDB needs testing

    def __init__(self, args):
        self.args = args
        self.cert_dir = self.log_dir = os.path.abspath(os.path.join(args.workdir, "certs"))
        if not os.path.isdir(self.cert_dir):
            os.makedirs(self.cert_dir)
        self.hash_fs = hashfs.HashFS(self.cert_dir, depth=4, width=1, algorithm='sha256')

    def put(self, der_data):
        if type(der_data) is str:
            hash_address = self.hash_fs.put(cStringIO.StringIO(str(der_data)), "der")
            logger.debug("Wrote certificate data to `%s`" % hash_address.abspath)
            return hash_address.id
        elif type(der_data) is list:
            return map(self.put, der_data)
        else:
            raise Exception("Unsupported argument type")

    def exists(self, hash_id):
        return self.hash_fs.exists(hash_id)

    def get_abspath(self, hash_id):
        hash_address = self.hash_fs.get(hash_id)
        if hash_address is None:
            return None
        else:
            return hash_address.abspath

    def get_relpath(self, hash_id):
        hash_address = self.hash_fs.get(hash_id)
        if hash_address is None:
            return None
        else:
            return hash_address.relpath


class RunLogDB(object):
    """
    Class to manage run log files
    """

    def __init__(self, args):
        self.args = args
        self.log_dir = os.path.abspath(os.path.join(args.workdir, "log"))
        if not os.path.isdir(self.log_dir):
            os.makedirs(self.log_dir)
        self.cert_db = CertDB(self.args)

    def handle_to_dir_name(self, handle):
        """
        Converts a log  handle to its directory name
        :param handle: str with handle
        :return: str with file name
        """
        # handle format is .strftime("%Y-%m-%dZ%H-%M-%S")
        year, month, _, _, _ = handle.split("-")
        return os.path.join(self.log_dir, year, month, handle)

    @staticmethod
    def dir_name_to_handle(dir_name):
        """
        Converts run log dir name to its handle
        :param dir_name: str
        :return: str with handle
        """
        return os.path.basename(dir_name)

    def exists(self, handle):
        """
        Check whether run log handle is valid
        :param handle: str with handle
        :return: bool
        """
        matches = glob.glob(os.path.join(self.log_dir, "2???", "??", "%s" % handle))
        return len(matches) > 0

    def list_logs(self):
        """
        Returns a list of available log directory names
        :return: list of str of available log directories
        """
        return glob.glob(os.path.join(self.log_dir, "2???", "??", "2???-??-??*"))

    def list(self):
        """
        Returns a list of log handles
        :return: list of str of log handles
        """
        return [self.dir_name_to_handle(dir_name) for dir_name in self.list_logs()]

    def delete(self, handle):
        """
        Delete a log specified by its handle. All associated
        files are purged from the database. There is no undo.
        :param handle: str with log handle
        :return: None
        """
        global logger
        # TODO: method needs testing
        dir_name = self.handle_to_dir_name(handle)
        logger.debug("Purging `%s` from run log database" % dir_name)
        shutil.rmtree(dir_name)

    def list_parts(self, handle):
        """
        Returns a list of parts for a log handles
        :return: list of str of log parts
        """
        # TODO: method needs testing
        return os.listdir(self.handle_to_dir_name(handle))

    def part_path(self, handle, part):
        """
        Return absolute path of a part inside a log's directory.
        :param handle: str with log handle
        :param part: str with part handle
        :return: str with absolute path
        """
        # TODO: method needs testing
        log_dir_path = self.handle_to_dir_name(handle)
        return os.path.join(log_dir_path, part)

    def open(self, handle, part, mode="r", compress=True):
        """
        Open a log file by handle and part name
        :param handle: str log handle
        :param part: str part name
        :param mode: str file mode
        :param compress: bool
        :return: file object
        """
        global logger

        part_name = self.part_path(handle, part)
        if "w" in mode and not os.path.isdir(os.path.dirname(part_name)):
            os.makedirs(os.path.dirname(part_name))

        if "r" in mode:
            if os.path.exists(part_name + ".bz2"):
                part_name = part_name + ".bz2"
        else:
            if compress:
                part_name += ".bz2"

        logger.debug("Opening run log file `%s` in mode `%s`" % (part_name, mode))

        try:
            if part_name.endswith(".bz2"):
                return bz2.BZ2File(part_name, mode)
            else:
                return open(part_name, mode)
        except IOError as err:
            raise Exception("Error opening run log file for handle `%s` part `%s`: %s" % (handle, part, err))

    def read(self, handle, part):
        """
        Return the string content of a log file referenced by its handle
        :param handle: str with handle
        :param part: str with part name
        :return: str with content of logfile
        """
        global logger
        logger.debug("Reading run log `%s` part `%s`" % (handle, part))
        with self.open(handle, part, "r") as f:
            return f.read().decode("utf-8")

    def write(self, handle, part, data):
        """
        Write to log file referenced by handle and part.
        The data object will be passed through str().encode("utf-8").
        :param handle: str with handle
        :param part: str with part name
        :param data: object to write
        :return: None
        """
        global logger
        logger.debug("Writing run log `%s`" % handle)
        with self.open(handle, part, "w") as f:
            f.write(str(data).encode("utf-8"))

    def put_cert(self, cert_data):
        """
        Add a certificate to the certificate database. Returns handle
        :param cert_data: str with DER data
        :return: str handle
        """
        # TODO: method needs testing
        return self.cert_db.put(cert_data)

    def new_log(self):
        """
        Return a new RunLog object that supports incremental logging
        :return:
        """
        handle = datetime.datetime.utcnow().strftime("%Y-%m-%dZ%H-%M-%S")
        return RunLog(handle, "w", self)

    def read_log(self, handle):
        """
        Return a RunLog object that works on the existing log
        refered to by handle.
        :param handle: str with handle
        :return: RunLog object
        """
        return RunLog(handle, "r", self)


class RunLog(object):
    """
    Class to keep state about runs in a generic format
    suitable for later processing. Calling the .log() method is
    designed to work incrementally such that we don't need to keep
    aggregating all results in memory.
    """

    format_revision = 2

    def __init__(self, handle, mode, db):
        """
        Constructor for RunLog

        The `mode` argument determines solely how the run log object
        behaves when called in a `with` statement. If the mode is "w",
        then .start() is called.

        Calling .start() switches the log to running mode. Iterating the
        log or calling len() while running throws an exception.

        :param handle: str with log handle
        :param mode: str with mode ("r" or "w")
        :param db: RunLogDB instance
        """
        self.db = db
        self.handle = handle
        self.mode = mode
        self.parts = []
        self.log_fh = None
        self.meta_fh = None
        self.filter = None
        self.meta = None
        self.is_running = False

    def part(self, part_handle):
        """
        Returns absolute path for an additional log part, referenced
        by its handle, inside the log directory.
        :param part_handle: str with handle
        :return: str with absolute path
        """
        # TODO: method needs testing
        return self.db.part_path(self.handle, part_handle)

    def open_part(self, part_handle, mode="r", compress=True):
        """
        Returns file object of given log part, openend in specified mode
        and compression. The class keeps internally track of file objects
        generated by this method, so they can all be closed at once.

        :param part_handle: str with handle
        :param mode: str with file mode
        :param compress: bool
        :return: file object
        """
        part_fh = self.db.open(self.handle, part_handle, mode, compress)
        self.parts.append(part_fh)
        return part_fh

    def close(self):
        """
        Close all files that were opened through .open_part().
        """
        if self.is_running:
            self.stop()

        if self.log_fh is not None:
            self.log_fh.close()
            self.log_fh = None
        if self.meta_fh is not None:
            self.meta_fh.close()
            self.meta_fh = None
        for f in self.parts:
            if not f.closed:
                f.close()
        self.parts = []

    def start(self, meta=None, log_filter=lambda x: x):
        """
        Start logging. Writes a metadata header to the file.

        Calling this method puts the log in running mode.
        Starting an existing log will overwrite log and meta data.

        :param meta: optional dict with metadata to write
        :param log_filter: filter lambda to map on every log line
        :return: None
        """
        if self.is_running:
            raise Exception("Can't restart already running log `%s`", self.handle)

        self.meta = meta if meta is not None else {}
        self.meta["format_revision"] = self.format_revision
        self.meta["run_completed"] = False
        self.meta["log_lines"] = 0
        self.filter = log_filter

        self.log_fh = self.open_part("log", self.mode, compress=True)
        self.meta_fh = self.open_part("meta", self.mode, compress=False)

        self.meta_fh.seek(0)
        self.meta_fh.write(json.dumps(self.meta, indent=4, sort_keys=True))
        self.meta_fh.truncate()

        self.is_running = True

    def log(self, result_batch):
        """
        Write an individual or multiple lines to the log file.
        Each line is passed through the filter function set by .start().
        A line is not logged if the filter returns None.
        :param result_batch: log item or list of log items
        :return: None
        """
        if not self.is_running:
            raise Exception("Unable to write to stopped log `%s`" % self.handle)

        if type(result_batch) is not list:
            result_batch = [result_batch]
        for result in map(self.filter, result_batch):
            if result is not None:
                self.log_fh.write("%s\n" % json.dumps(result))
                self.meta["log_lines"] += 1

    def stop(self, meta=None):
        """
        Wrap up the log file. The given metadata is joined with previously
        provided metadata and appended at the end of the file. Finally, an
        indicator is written to the log file that is used for distinguishing
        between complete and partial log files.

        Metadata given to .stop() has precedence over metadata given to .start().
        :param meta: dict with additional metadata
        :return: None
        """
        if not self.is_running:
            raise Exception("Unable to stop stopped log `%s`", self.handle)

        if meta is None:
            meta = {}
        self.meta.update(meta)
        self.meta["run_completed"] = True
        self.meta_fh.seek(0)
        self.meta_fh.write(json.dumps(self.meta, indent=4, sort_keys=True))
        self.meta_fh.truncate()

        self.meta_fh.close()
        self.meta_fh = None
        self.log_fh.close()
        self.log_fh = None

        self.is_running = False

    def has_finished(self):
        """
        Checks whether the run finished. If the log file is in write mode,
        it checks whether the .stop() method has been called, yet. If the log is in
        read mode, it checks for the presence of the `run_completed` metadata.
        :return: bool
        """
        if self.is_running:
            return False
        meta = self.get_meta()
        if "run_completed" not in meta:
            return False
        else:
            return meta["run_completed"]

    def is_compatible(self):
        """
        Returns whether the log has the same revision as RunLog.format_revision.
        :return: bool
        """
        if self.is_running:
            return True
        meta = self.get_meta()
        if "format_revision" not in meta:
            return False
        else:
            return meta["format_revision"] == self.format_revision

    def update_meta(self, meta):
        """
        Update the metadata on a log. If the log is running, it will not be
        written until .stop() is called.

        Keep in mind that metadata is only updated with the data provided.
        Keys can't be deleted.
        :param meta: dict
        :return: None
        """
        if self.is_running:
            self.meta.update(meta)
        else:
            self.get_meta()
            self.meta.update(meta)
            with self.open_part("meta", "w") as f:
                f.write(json.dumps(self.meta, indent=4, sort_keys=True))

    def get_meta(self):
        """
        Get metadata from log. If the log is running, the metadata collected
        by calls to .start() and .stop() is returned. If the log is not running,
        metadata is taken from the `meta` property of this object.
        :return: dict with metadata
        """
        if self.is_running:
            return self.meta
        if self.meta is None:
            with self.open_part("meta") as f:
                meta = f.read().strip()
            try:
                self.meta = json.loads(meta)
            except ValueError:
                self.meta = {"CORRUPTED": True}
        return self.meta

    def __iter__(self):
        """
        If the log is in read mode, iterate over all log lines.
        :return: iterator
        """
        if self.is_running:
            raise Exception("Unable to iterate running log `%s`", self.handle)

        with self.open_part("log") as f:
            line_number = 0
            while True:
                try:
                    line_number += 1
                    line = f.readline()
                except EOFError:
                    logger.debug("EOFError on log `%s`. Log is truncated." % self.handle)
                    break
                if line == "":
                    break
                if not line.startswith("#"):
                    try:
                        json_line = json.loads(line.strip())
                    except ValueError as err:
                        raise Exception("JSON error in run log `%s` line %d: %s"
                                        % (self.handle, line_number, str(err)))
                    yield json_line

    def __len__(self):
        """
        Return the number of lines in the log.

        If the log is running, returns the number of lines written so far.
        If the log is not running and was completed, it returns the value of
        `log_lines` from the metadata.
        If the log is not running and was not completed, it iterates all the
        log lines to count determine the number (slow!).
        """
        # TODO: method needs testing
        if self.is_running:
            return self.meta["log_lines"]

        if self.has_finished():
            return self.get_meta()["log_lines"]

        logger.debug("Counting lines in incomplete log `%s`" % self.handle)
        incomplete_log_lines = 0
        for _ in self:
            incomplete_log_lines += 1
        return incomplete_log_lines

    def __enter__(self):
        """Called when object enters a `with` block"""
        if "w" in self.mode:
            self.start()
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        """Called when object exits a `with` block"""
        if self.is_running:
            self.stop()
        self.close()
        if exception_type is not None:
            logger.error("Exception while handling run log %s: %s %s %s"
                         % (self.handle, exception_type, exception_value, traceback))
        return self

    def delete(self):
        """
        Delete this log from the log database. All associated files are
        purged from disk. Be careful! There is no undo.

        After calling this method, the log object becomes dysfunctional.
        No other methods should be called hereafter.
        :return: None
        """
        # TODO: method needs testing
        if self.is_running:
            self.stop()
        self.close()
        self.db.delete(self.handle)

    def put_cert(self, cert_data):
        """
        Add a certificate to the certificate database. Return handle
        :param cert_data: str with DER data
        :return: str handle
        """
        # TODO: method needs testing
        return self.db.put_cert(cert_data)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import logging
import os
import struct
import sys
import urllib2

import cache


logger = logging.getLogger(__name__)


def get_to_file(url, filename):
    global logger

    try:
        # TODO: Validate the server's SSL certificate
        req = urllib2.urlopen(url)
        file_size = int(req.info().get('Content-Length').strip())

        # Caching logic is: don't re-download if file of same size is
        # already in place. TODO: Switch to ETag if that's not good enough.
        # This already prevents cache clutter with incomplete files.
        if os.path.isfile(filename):
            if os.stat(filename).st_size == file_size:
                req.close()
                logger.warning('Skipping download, using cached file `%s` instead' % filename)
                return filename
            else:
                logger.warning('Purging incomplete or obsolete cache file `%s`' % filename)
                os.remove(filename)

        logger.debug('Downloading `%s` to %s' % (url, filename))
        downloaded_size = 0
        chunk_size = 32 * 1024
        with open(filename, 'wb') as fp:
            while True:
                chunk = req.read(chunk_size)
                if not chunk:
                    break
                downloaded_size += len(chunk)
                fp.write(chunk)

    except urllib2.HTTPError, err:
        if os.path.isfile(filename):
            os.remove(filename)
        logger.error('HTTP error: %s, %s' % (err.code, url))
        return None

    except urllib2.URLError, err:
        if os.path.isfile(filename):
            os.remove(filename)
        logger.error('URL error: %s, %s' % (err.reason, url))
        return None

    except KeyboardInterrupt:
        if os.path.isfile(filename):
            os.remove(filename)
        if sys.stdout.isatty():
            print
        logger.critical('Download interrupted by user')
        return None

    return filename


class FirefoxDownloader(object):

    __base_url = 'https://download.mozilla.org/?product=firefox' \
                '-{release}&os={platform}&lang=en-US'
    build_urls = {
        'esr':     __base_url.format(release='esr-latest', platform='{platform}'),
        'release': __base_url.format(release='latest', platform='{platform}'),
        'beta':    __base_url.format(release='beta-latest', platform='{platform}'),
        'aurora':  __base_url.format(release='aurora-latest', platform='{platform}'),
        'nightly': __base_url.format(release='nightly-latest', platform='{platform}')
    }
    __platforms = {
        'osx':     {'platform': 'osx', 'extension': 'dmg'},
        'linux':   {'platform': 'linux64', 'extension': 'tar.bz2'},
        'linux32': {'platform': 'linux', 'extension': 'tar.bz2'},
        'win':     {'platform': 'win64', 'extension': 'exe'},
        'win32':   {'platform': 'win', 'extension': 'exe'}
    }

    @staticmethod
    def list():
        build_list = FirefoxDownloader.build_urls.keys()
        platform_list = FirefoxDownloader.__platforms.keys()
        test_default = "nightly"
        base_default = "release"
        return build_list, platform_list, test_default, base_default

    @staticmethod
    def detect_platform():
        is_64bit = struct.calcsize('P') * 8 == 64
        platform = None
        if sys.platform.startswith("darwin"):
            platform = "osx"
        elif sys.platform.startswith("linux"):
            platform = "linux" if is_64bit else "linux32"
        elif sys.platform.startswith("win"):
            platform = "win" if is_64bit else "win32"
        return platform

    @staticmethod
    def get_download_url(build, platform=None):
        if platform is None:
            platform = FirefoxDownloader.detect_platform()
        # We internally use slightly different platform naming, so translate
        # internal platform name to the platform name used in download URL.
        download_platform = FirefoxDownloader.__platforms[platform]['platform']
        if build in FirefoxDownloader.build_urls:
            return FirefoxDownloader.build_urls[build].format(platform=download_platform)
        else:
            return None

    def __init__(self, workdir, cache_timeout=24*60*60):
        self.__workdir = workdir
        self.__cache = cache.DiskCache(os.path.join(workdir, "cache"), cache_timeout, purge=True)

    def download(self, release, platform=None, use_cache=True):

        if platform is None:
            platform = self.detect_platform()

        if release not in self.build_urls:
            raise Exception("Failed to download unknown release `%s`" % release)
        if platform not in self.__platforms:
            raise Exception("Failed to download for unknown platform `%s`" % platform)

        extension = self.__platforms[platform]['extension']
        url = self.get_download_url(release, platform)
        cache_id = 'firefox-%s_%s.%s' % (release, platform, extension)

        # Always delete cached file when cache function is overridden
        if cache_id in self.__cache and not use_cache:
            self.__cache.delete(cache_id)

        # __get_to_file will not re-download if same-size file is already there.
        return get_to_file(url, self.__cache[cache_id])



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import datetime
import logging
import pkg_resources as pkgr
import sys

from basemode import BaseMode
import tlscanary.progress as pr
import tlscanary.runlog as rl
import tlscanary.sources_db as sdb


logger = logging.getLogger(__name__)


class ScanMode(BaseMode):

    name = "scan"
    help = "Collect SSL connection state info on hosts"

    def __init__(self, args, module_dir, tmp_dir):
        global logger

        super(ScanMode, self).__init__(args, module_dir, tmp_dir)

        # Define instance attributes for later use
        self.log = None
        self.start_time = None
        self.test_profile = None
        self.test_app = None
        self.test_metadata = None
        self.sources = None

    def setup(self):
        global logger

        # Argument validation logic to make sure user has specified only test build
        if self.args.test is None:
            logger.critical("Must specify test build for scan")
            sys.exit(5)
        elif self.args.base is not None:
            logger.debug("Ignoring base build parameter")

        # Download app and extract metadata
        self.test_app = self.get_test_candidate(self.args.test)
        self.test_metadata = self.collect_worker_info(self.test_app)
        logger.info("Testing Firefox %s %s scan run" %
                    (self.test_metadata["app_version"], self.test_metadata["branch"]))

        # Create custom profile
        self.test_profile = self.make_profile("test_profile")

        # Compile the set of hosts to test
        db = sdb.SourcesDB(self.args)
        logger.info("Reading `%s` host database" % self.args.source)
        self.sources = db.read(self.args.source)
        logger.info("%d hosts in test set" % len(self.sources))

    def run(self):
        global logger

        logger.info("Testing Firefox %s %s" %
                    (self.test_metadata["app_version"], self.test_metadata["branch"]))

        self.start_time = datetime.datetime.now()

        meta = {
            "tlscanary_version": pkgr.require("tlscanary")[0].version,
            "mode": self.name,
            "args": vars(self.args),
            "argv": sys.argv,
            "sources_size": len(self.sources),
            "test_metadata": self.test_metadata,
            "run_start_time": datetime.datetime.utcnow().isoformat()
        }

        rldb = rl.RunLogDB(self.args)
        log = rldb.new_log()
        log.start(meta=meta)
        progress = pr.ProgressTracker(total=len(self.sources), unit="hosts", average=30*60.0)

        limit = len(self.sources) if self.args.limit is None else self.args.limit

        # Split work into 50 chunks to conserve memory, but make no chunk smaller than 1000 hosts
        next_chunk = self.sources.iter_chunks(chunk_size=limit/50, min_chunk_size=1000)

        try:
            while True:
                host_set_chunk = next_chunk(as_set=True)
                if host_set_chunk is None:
                    break

                logger.info("Starting scan of chunk of %d hosts" % len(host_set_chunk))

                info_uri_set = self.run_test(self.test_app, host_set_chunk, profile=self.test_profile,
                                             prefs=self.args.prefs, get_info=True,
                                             get_certs=not self.args.remove_certs, return_only_errors=False,
                                             report_callback=progress.log_completed)
                # Log progress per chunk
                logger.info("Progress: %s" % str(progress))

                # Commit results to log
                for rank, host, result in info_uri_set:
                    log.log(result.as_dict())

        except KeyboardInterrupt:
            logger.critical("Ctrl-C received")
            progress.stop_reporting()
            raise KeyboardInterrupt

        finally:
            progress.stop_reporting()

        meta["run_finish_time"] = datetime.datetime.utcnow().isoformat()
        self.save_profile(self.test_profile, "test_profile", log)
        log.stop(meta=meta)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import atexit
import signal
import sys


__cleanup_done = False


def init():
    """Register cleanup handler"""

    # print "Registering cleanup handler"

    global __cleanup_done
    __cleanup_done = False

    # Will be OS-specific, see https://docs.python.org/2/library/signal.html
    atexit.register(cleanup_handler)
    signal.signal(signal.SIGTERM, cleanup_handler)
    if sys.platform == "darwin" or "linux" in sys.platform:
        # SIGHUP is not available on Windows
        signal.signal(signal.SIGHUP, cleanup_handler)


def cleanup_handler():
    """The cleanup handler that runs when process terminates"""
    # print "Cleanup handler called"
    global __cleanup_done
    if not __cleanup_done:
        __cleanup_done = True
        for child in CleanUp.__subclasses__():
            child.at_exit()


class CleanUp(object):
    """When process terminates, .at_exit() is called on every subclass."""
    pass



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.


from distutils import dir_util
import glob
import logging
import os
import shutil
import stat
import sys
import zipfile

import tlscanary.firefox_app as fa
import tlscanary.firefox_downloader as fd
import tlscanary.firefox_extractor as fe
import tlscanary.one_crl_downloader as one_crl
import tlscanary.sources_db as sdb
import tlscanary.worker_pool as wp
import tlscanary.xpcshell_worker as xw


logger = logging.getLogger(__name__)


class BaseMode(object):
    """
    Generic Test Mode
    Base functionality for all tests
    """

    @classmethod
    def setup_args(cls, parser):
        """
        Add subparser for the mode's specific arguments.

        This definition serves as default, but modes are free to
        override it.

        :param parser: parent argparser to add to
        :return: None
        """

        # By nature of workdir being undetermined at this point, user-defined test sets in
        # the override directory can not override the default test set. The defaulting logic
        # needs to move behind the argument parser for that to happen.
        src = sdb.SourcesDB()
        testset_default = src.default
        release_choice, _, test_default, base_default = fd.FirefoxDownloader.list()

        group = parser.add_argument_group("test candidates selection")
        group.add_argument('-t', '--test',
                           help=("Firefox version to test. It can be one of {%s}, a package file, "
                                 "or a build directory (default: `%s`)") % (",".join(release_choice), test_default),
                           action='store',
                           default=test_default)
        group.add_argument("-b", "--base",
                           help=("Firefox base version to compare against. It can be one of {%s}, a package file, "
                                 "or a build directory (default: `%s`)") % (",".join(release_choice), base_default),
                           action='store',
                           default=base_default)

        group = parser.add_argument_group("profile setup")
        group.add_argument("-c", "--cache",
                           help='Allow profiles to cache web content',
                           action="store_true",
                           default=False)
        group.add_argument("-o", "--onecrl",
                           help="OneCRL set to test (default: production)",
                           type=str.lower,
                           choices=["production", "stage", "custom", "none"],
                           action="store",
                           default="production")
        group.add_argument("--onecrlpin",
                           help="OneCRL-Tools git branch, tag, or commit to use (default: stable)",
                           action="store",
                           default=os.environ.get("ONECRLPIN", "stable"))
        group.add_argument("-p", "--prefs",
                           help="Prefs to apply to all builds",
                           type=str,
                           action="append",
                           default=None)
        group.add_argument("-p1", "--prefs_test",
                           help="Prefs to apply to test build",
                           type=str,
                           action="append",
                           default=None)
        group.add_argument("-p2", "--prefs_base",
                           help="Prefs to apply to base build",
                           type=str,
                           action="append",
                           default=None)

        group = parser.add_argument_group("host database selection")
        group.add_argument("-s", "--source",
                           help="Test set to run. Use `list` for info. (default: `%s`)" % testset_default,
                           action="store",
                           default=testset_default)
        group.add_argument("-l", "--limit",
                           help="Limit for number of hosts to test (default: no limit)",
                           type=int,
                           action="store",
                           default=None)

        group = parser.add_argument_group("worker configuration")
        group.add_argument("-j", "--parallel",
                           help="Number of parallel worker instances (default: 4)",
                           type=int,
                           action="store",
                           default=4)
        group.add_argument("-m", "--timeout",
                           help="Timeout for worker requests (default: 10)",
                           type=float,
                           action="store",
                           default=10)
        group.add_argument("-n", "--requestsperworker",
                           help="Number of requests per worker (default: 50)",
                           type=int,
                           action="store",
                           default=50)
        group.add_argument("-u", "--max_timeout",
                           help="Maximum timeout for worker requests (default: 20)",
                           type=float,
                           action="store",
                           default=20)
        group.add_argument("-x", "--scans",
                           help="Number of scans per host (default: 3)",
                           type=int,
                           action="store",
                           default=3)

        group = parser.add_argument_group("result processing")
        group.add_argument("-f", "--filter",
                           help="Filter level for results 0:none 1:timeouts (default: 1)",
                           type=int,
                           choices=[0, 1],
                           action="store",
                           default=1)
        group.add_argument("-r", "--remove_certs",
                           help="Filter certificate data from results",
                           action="store_true")

    def __init__(self, args, module_dir, tmp_dir):
        self.args = args
        self.mode = args.mode
        self.module_dir = module_dir
        self.tmp_dir = tmp_dir

    def get_test_candidate(self, build):
        """
        Download and extract a build candidate. build may either refer
        to a Firefox release identifier, package, or build directory.
        :param build: str with firefox build
        :return: two FirefoxApp objects for test and base candidate
        """
        global logger

        platform = fd.FirefoxDownloader.detect_platform()
        if platform is None:
            logger.error("Unsupported platform: `%s`" % sys.platform)
            sys.exit(5)

        # `build` may refer to a build reference as defined in FirefoxDownloader,
        # a local Firefox package as produced by `mach build`, or a local build tree.
        if build in fd.FirefoxDownloader.build_urls:
            # Download test candidate by Firefox release ID
            logger.info("Downloading Firefox `%s` build for platform `%s`" % (build, platform))
            fdl = fd.FirefoxDownloader(self.args.workdir, cache_timeout=1 * 60 * 60)
            build_archive_file = fdl.download(build, platform)
            if build_archive_file is None:
                sys.exit(-1)
            # Extract candidate archive
            candidate_app = fe.extract(build_archive_file, self.args.workdir, cache_timeout=1 * 60 * 60)
            candidate_app.package_origin = fdl.get_download_url(build, platform)
        elif os.path.isfile(build):
            # Extract firefox build from archive
            logger.info("Using file `%s` as Firefox package" % build)
            candidate_app = fe.extract(build, self.args.workdir, cache_timeout=1 * 60 * 60)
            candidate_app.package_origin = build
            logger.debug("Build candidate executable is `%s`" % candidate_app.exe)
        elif os.path.isfile(os.path.join(build, "mach")):
            logger.info("Using Firefox build tree at `%s`" % build)
            dist_globs = sorted(glob.glob(os.path.join(build, "obj-*", "dist")))
            if len(dist_globs) == 0:
                logger.critical("`%s` looks like a Firefox build directory, but can't find a build in it" % build)
                sys.exit(5)
            logger.debug("Potential globs for dist directory: %s" % dist_globs)
            dist_dir = dist_globs[-1]
            logger.info("Using `%s` as build distribution directory" % dist_dir)
            if "apple-darwin" in dist_dir.split("/")[-2]:
                # There is a special case for OS X dist directories:
                # FirefoxApp expects OS X .dmg packages to contain the .app folder inside
                # another directory. However, that directory isn't there in build trees,
                # thus we need to point to the parent for constructing the app.
                logger.info("Looks like this is an OS X build tree")
                candidate_app = fa.FirefoxApp(os.path.abspath(os.path.dirname(dist_dir)))
                candidate_app.package_origin = os.path.abspath(build)
            else:
                candidate_app = fa.FirefoxApp(os.path.abspath(dist_dir))
                candidate_app.package_origin = os.path.abspath(build)
        else:
            logger.critical("`%s` specifies neither a Firefox release, package file, or build directory" % build)
            logger.critical("Valid Firefox release identifiers are: %s" % ", ".join(fd.FirefoxDownloader.list()[0]))
            sys.exit(5)

        logger.debug("Build candidate executable is `%s`" % candidate_app.exe)
        if candidate_app.platform != platform:
            logger.warning("Platform mismatch detected")
            logger.critical("Running a Firefox binary for `%s` on a `%s` platform will probably fail" %
                            (candidate_app.platform, platform))
        return candidate_app

    @staticmethod
    def collect_worker_info(app):
        worker = xw.XPCShellWorker(app)
        worker.spawn()
        worker.send(xw.Command("info"))
        result = worker.wait().result
        worker.terminate()
        # Add convenience shortcuts and metadata not returned by worker
        result["nss_version"] = "NSS %s" % result["nssInfo"]["NSS_Version"]
        result["nspr_version"] = "NSPR %s" % result["nssInfo"]["NSPR_Version"]
        result["branch"] = result["appConstants"]["MOZ_UPDATE_CHANNEL"]
        result["app_version"] = result["appConstants"]["MOZ_APP_VERSION_DISPLAY"]
        result["application_ini"] = app.application_ini
        result["package_origin"] = app.package_origin
        result["platform"] = app.platform
        return result

    def make_profile(self, profile_name, one_crl_env='production'):
        global logger

        # create directories for profiles
        default_profile_dir = os.path.join(self.module_dir, "default_profile")
        new_profile_dir = os.path.join(self.tmp_dir, profile_name)

        if not os.path.exists(new_profile_dir):
            os.makedirs(new_profile_dir)

        # copy contents of default profile to new profiles
        dir_util.copy_tree(default_profile_dir, new_profile_dir)

        logger.info("Updating OneCRL revocation data")
        if one_crl_env == "production" or one_crl_env == "stage":
            # overwrite revocations file in test profile with live OneCRL entries from requested environment
            revocations_file = one_crl.get_list(one_crl_env, self.args.workdir, self.args.onecrlpin)
            profile_file = os.path.join(new_profile_dir, "revocations.txt")
            logger.debug("Writing OneCRL revocations data to `%s`" % profile_file)
            shutil.copyfile(revocations_file, profile_file)
        elif one_crl_env == "none":
            # delete revocations file completely
            logger.debug("Deleting existing revocations.txt file")
            profile_file = os.path.join(new_profile_dir, "revocations.txt")
            os.remove(profile_file)
        else:
            # leave the existing revocations file alone
            logger.info("Testing with custom OneCRL entries from default profile")

        logger.debug("Allow profile cache: %s" % self.args.cache)
        if not self.args.cache:
            # make all files in profiles read-only to prevent caching
            for root, dirs, files in os.walk(new_profile_dir, topdown=False):
                for name in files:
                    os.chmod(os.path.join(root, name), stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)

        return new_profile_dir

    @staticmethod
    def save_profile(profile_path, profile_name, log):
        global logger

        log_part = "%s.zip" % profile_name
        zip_file_path = log.part(log_part)
        logger.debug("Saving `%s` profile from `%s` to `%s`" % (profile_name, profile_path, zip_file_path))

        if not log.is_running:
            logger.critical("Can't save profile to closed log. Please file a bug.")
            return

        with zipfile.ZipFile(zip_file_path, "w", zipfile.ZIP_DEFLATED) as z:
            for root, _, files in os.walk(profile_path):
                for f in files:
                    file_name = os.path.join(root, f)
                    arc_name = os.path.relpath(file_name, profile_path)
                    z.write(file_name, arc_name)

        if "profiles" not in log.meta:
            log.meta["profiles"] = []
        log.meta["profiles"].append({"name": profile_name, "log_part": log_part})

    def run_test(self, app, url_list, profile=None, prefs=None, num_workers=None, n_per_worker=None, timeout=None,
                 get_info=False, get_certs=False, return_only_errors=True, report_callback=None):

        global logger

        # Default to values from args
        if num_workers is None:
            num_workers = self.args.parallel
        if n_per_worker is None:
            n_per_worker = self.args.requestsperworker
        if timeout is None:
            timeout = self.args.timeout

        try:
            results = wp.run_scans(app, list(url_list), profile=profile, prefs=prefs, num_workers=num_workers,
                                   targets_per_worker=n_per_worker, timeout=timeout,
                                   get_certs=get_certs, progress_callback=report_callback)

        except KeyboardInterrupt:
            logger.critical('User abort')
            wp.stop()
            sys.exit(1)

        run_results = set()

        for host in results:
            if return_only_errors:
                if not results[host].success:
                    if get_info:
                        run_results.add((results[host].rank, host, results[host]))
                    else:
                        run_results.add((results[host].rank, host))
            else:
                if get_info:
                    run_results.add((results[host].rank, host, results[host]))
                else:
                    run_results.add((results[host].rank, host))

        return run_results

    def setup(self):
        """
        Runs all the steps required before doing the mode runs.
        Put everything here that takes too long for __init__().
        :return: None
        """
        pass

    def run(self):
        """
        Executes the the steps that constitutes the actual mode run.
        Results are kept internally in the class instance.
        :return: None
        """
        pass

    def teardown(self):
        """
        Clean up steps required after a mode run.
        :return: None
        """
        pass



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import json
import logging
import os
from Queue import Queue, Empty
import subprocess
from threading import Thread


logger = logging.getLogger(__name__)
module_dir = os.path.split(__file__)[0]


def read_from_worker(worker, response_queue):
    """Reader thread that reads messages from the worker.
       The convention is that all worker output that parses
       as JSON is routed to the response queue, else it is
       interpreted as a JavaScript error or warning.
    """
    global logger

    logger.debug('Reader thread started for worker %s' % worker)
    for line in iter(worker.stdout.readline, b''):
        line = line.strip()
        try:
            response_queue.put(Response(line))
            logger.debug("Received worker message: %s" % line)
        except ValueError:
            if line.startswith("JavaScript error:"):
                logger.error("JS error from worker %s: %s" % (worker, line))
            elif line.startswith("JavaScript warning:"):
                logger.warning("JS warning from worker %s: %s" % (worker, line))
            else:
                logger.critical("Invalid output from worker %s: %s" % (worker, line))
    logger.debug('Reader thread finished for worker %s' % worker)
    worker.stdout.close()


class XPCShellWorker(object):
    """XPCShell worker implementing an asynchronous, JSON-based message system"""

    def __init__(self, app, script=None, profile=None, prefs=None):
        global module_dir

        self.__app = app
        if script is None:
            self.__script = os.path.join(module_dir, "js", "scan_worker.js")
        else:
            self.__script = script
        self.__profile = profile
        self.__prefs = prefs
        self.__worker_thread = None
        self.__reader_thread = None
        self.__response_queue = Queue()

    def spawn(self):
        """Spawn the worker process and its dedicated reader thread"""
        global logger, module_dir

        cmd = [self.__app.exe, '-xpcshell', "-g", self.__app.gredir, "-a", self.__app.browser, self.__script]
        logger.debug("Executing worker shell command `%s`" % ' '.join(cmd))

        self.__worker_thread = subprocess.Popen(
            cmd,
            cwd=self.__app.browser,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            bufsize=1)  # `1` means line-buffered

        # Spawn a reader thread, because stdio reads are blocking
        self.__reader_thread = Thread(target=read_from_worker, name="Reader",
                                      args=(self.__worker_thread, self.__response_queue))
        self.__reader_thread.daemon = True  # Thread dies with worker
        self.__reader_thread.start()

        if self.__profile is not None:
            logger.debug("Changing worker profile to `%s`" % self.__profile)
            if self.send(Command("useprofile", path=self.__profile)):
                response = self.wait()
            else:
                # .wait() would wait forever if .send() was not successful
                response = None
            if response is None or response.original_cmd["mode"] != "useprofile" or response.result != "ACK":
                logger.error("Worker failed to set profile `%s`" % self.__profile)
                return False

        if self.__prefs is not None:
            logger.debug("Setting worker prefs to `%s`" % self.__prefs)
            if self.send(Command("setprefs", prefs=self.__prefs)):
                response = self.wait()
            else:
                # .wait() would wait forever if .send() was not successful
                response = None
            if response is None or response.original_cmd["mode"] != "setprefs" or response.result != "ACK":
                logger.error("Worker failed to set prefs `%s`" % self.__prefs)
                return False

        return True

    def terminate(self):
        """Signal the worker process to quit"""
        # The reader thread dies when the Firefox process quits
        self.__worker_thread.terminate()

    def kill(self):
        """Kill the worker process"""
        self.__worker_thread.kill()

    def is_running(self):
        """Check whether the worker is still fully running"""
        if self.__worker_thread is None:
            return False
        return self.__worker_thread.poll() is None

    def send(self, cmd):
        """Send a command message to the worker"""
        global logger

        cmd_string = str(cmd)
        logger.debug("Sending worker message: `%s`" % cmd_string)
        try:
            self.__worker_thread.stdin.write((cmd_string + "\n").encode("utf-8"))
            self.__worker_thread.stdin.flush()
        except IOError:
            logger.debug("Can't write to worker. Message `%s` wasn't heard." % cmd_string)
            return False
        return True

    def receive(self):
        """Read queued messages from worker. Returns [] if there were none."""

        global logger

        # Read everything from the reader queue
        responses = []
        try:
            while True:
                responses.append(self.__response_queue.get_nowait())
        except Empty:
            pass

        return responses

    def wait(self):
        """Wait for and return the next single message from the worker."""
        return self.__response_queue.get()


class Command(object):

    def __init__(self, mode, id=None, **kwargs):
        if mode is None:
            raise Exception("Refusing to init mode-less command")
        self.__id = id
        self.__mode = mode
        self.__args = kwargs

    def as_dict(self):
        return {"id": self.__id, "mode": self.__mode, "args": self.__args}

    def __str__(self):
        return json.dumps(self.as_dict())


class Response(object):

    def __init__(self, message_string):
        global logger

        self.id = None
        self.worker_id = None
        self.original_cmd = None
        self.success = None
        self.result = None
        self.elapsed_ms = None
        message = json.loads(message_string)  # May throw ValueError
        if "id" in message:
            self.id = message["id"]
        if "original_cmd" in message:
            self.original_cmd = message["original_cmd"]
        if "worker_id" in message:
            self.worker_id = message["worker_id"]
        if "success" in message:
            self.success = message["success"]
        if "result" in message:
            self.result = message["result"]
        if "command_time" in message:
            self.command_time = message["command_time"]
        if "response_time" in message:
            self.response_time = message["response_time"]
        if len(message) != 7:
            logger.error("Worker response has unexpected format: %s" % message_string)

    def as_dict(self):
        return {
            "id": self.id,
            "original_cmd": self.original_cmd,
            "worker_id": self.worker_id,
            "success": self.success,
            "result": self.result,
            "command_time": self.command_time,
            "response_time": self.response_time,
        }



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import csv
import logging
import os
import pkg_resources as pkgr


logger = logging.getLogger(__name__)


def list_sources(override_dir=None):
    """
    This function trawls through all the sources CSV files in the module CSV directory and
    the given override directory, and generates a dictionary of database handle names and
    associated file names. Per default, the base part of the file name (without `.csv`) is
    used as handle for that list.

    Files in the override directory override files in the module directory.

    If the first line of a CSV file begins with a `#`, it is interpreted as a
    colon-separated list of keywords. If it contains the keyword `handle`, the last
    keyword is used as its handle instead of the file name derivative.

    If the line contains the keyword `default`, it is being used as the default list.
    When multiple CSV files use the `default` keyword, the lexicographically last file
    name is used as default.

    :param override_dir: str of directory used for overrides
    :return: (dict mapping handles to file names, str handle of default list)
    """
    global logger

    sources_list = {}
    default_source = None

    # First, look for CSV files in module resources
    csv_files = [os.path.abspath(pkgr.resource_filename(__name__, "sources/%s" % name))
                 for name in pkgr.resource_listdir(__name__, "sources")
                 if name.endswith(".csv")]

    # Then look for CSV files in the override directory
    if override_dir is not None and os.path.isdir(override_dir):
        for root, dirs, files in os.walk(override_dir):
            for name in files:
                if name.endswith(".csv"):
                    csv_files.append(os.path.abspath(os.path.join(root, name)))

    # Finally extract metadata from files and compile sources list
    for file_name in csv_files:
        logger.debug("Indexing database resource `%s`" % file_name)
        source_handle, is_default = parse_csv_header(file_name)
        sources_list[source_handle] = file_name
        if is_default:
            default_source = source_handle

    return sources_list, default_source


def parse_csv_header(file_name):
    """
    Read first line of file and try to interpret it as a series of colon-separated
    keywords if the line starts with a `#`. Currently supported keywords:

    - handle: The last keyword is interpreted as database vanity handle
    - default: The database is used as default database.

    If no handle is specified, the file name's base is used instead.

    :param file_name: str with file name to check
    :return: (string with handle, bool default state)
    """
    source_handle = os.path.splitext(os.path.basename(file_name))[0]
    is_default = False
    with open(file_name) as f:
        line = f.readline().strip()
    if line.startswith("#"):
        keywords = line.lstrip("#").split(":")
        if "handle" in keywords:
            source_handle = keywords[-1]
        if "default" in keywords:
            is_default = True
    return source_handle, is_default


class SourcesDB(object):
    """
    Class to represent the database store for host data. CSV files from the `sources`
    subdirectory of the module directory are considered as database source files.
    Additionally, CSV files inside the `sources` subdirectory of the working directory
    (usually ~/.tlscanary) are parsed and thus can override files from the module
    directory.

    Each database file is referenced by a unique handle. The first line of the CSV can
    be a special control line that modifies how the database file is handled. See
    sources_db.parse_csv_header().

    The CSV files are required to contain a  regular CSV header line, the column
    `hostname`, and optionally the column `rank`.
    """
    def __init__(self, args=None):
        self.__args = args
        if args is not None:
            self.__override_dir = os.path.join(args.workdir, "sources")
        else:
            self.__override_dir = None
        self.__list, self.default = list_sources(self.__override_dir)
        if self.default is None:
            self.default = self.__list.keys()[0]

    def list(self):
        """
        List handles of available source CSVs

        :return: list with handles
        """
        handles_list = self.__list.keys()
        handles_list.sort()
        return handles_list

    def read(self, handle):
        """
        Read the database file referenced by the given handle.

        :param handle: str with handle
        :return: Sources object containing the data
        """
        global logger
        if handle not in self.__list:
            logger.error("Unknown sources database handle `%s`. Continuing with empty set" % handle)
            return Sources(handle)
        file_name = self.__list[handle]
        source = Sources(handle, handle == self.default)
        source.load(file_name)
        source.trim(self.__args.limit)
        return source

    def write(self, source):
        """
        Write a Sources object to a CSV database file into the `sources` subdirectory of
        the working directory (usually ~/.tlscanary). The file is named <handle.csv>.
        Metadata like handle and default state are stored in the first line of the file.

        :param source: Sources object
        :return: None
        """
        sources_dir = os.path.join(self.__args.workdir, "sources")
        if not os.path.isdir(sources_dir):
            os.makedirs(sources_dir)
        file_name = os.path.join(sources_dir, "%s.csv" % source.handle)
        source.write(file_name)


class Sources(object):
    def __init__(self, handle, is_default=False):
        self.handle = handle
        self.is_default = is_default
        self.rows = []
        # State for chunked iteration
        self.chunk_start = None
        self.chunk_stop = None
        self.chunk_size = None
        self.chunk_offset = None

    def __len__(self):
        return len(self.rows)

    def __getitem__(self, item):
        return self.rows[item]

    def __iter__(self):
        for row in self.rows:
            yield row

    def append(self, row):
        """
        Add a row to the end of the current sources list

        :param row: dict of `rank` and `hostname`
        :return:  None
        """
        self.rows.append(row)

    def sort(self):
        """
        Sort rows according to rank

        :return: None
        """
        self.rows.sort(key=lambda row: int(row["rank"]))

    def load(self, file_name):
        """
        Load content of a sources database from a CSV file

        :param file_name: str containing existing file name
        :return: None
        """
        global logger
        self.handle, self.is_default = parse_csv_header(file_name)
        logger.debug("Reading `%s` sources from `%s`" % (self.handle, file_name))
        with open(file_name) as f:
            csv_reader = csv.DictReader(filter(lambda r: not r.startswith("#"), f))
        self.rows = [row for row in csv_reader]

    def trim(self, limit):
        """
        Trim length of sources list to given limit. Does not trim if
        limit is None.

        :param limit: int maximum length or None
        :return: None
        """
        if limit is not None:
            if len(self) > limit:
                self.rows = self.rows[:limit]

    def write(self, location):
        """
        Write out instance sources list to a CSV file. If location refers to
        a directory, the file is written there and the file name is chosen as
        <handle>.csv. Metadata like handle and default state are stored in the
        first line of the file.

        If location refers to a file name, it used as file name directly.
        The target directory must exist.

        :param location: directory or file name in an existing directory
        :return: None
        """
        global logger
        if os.path.isdir(location):
            file_name = os.path.join(location, "%s.csv" % self.handle)
        elif os.path.isdir(os.path.dirname(location)):
            file_name = location
        else:
            raise Exception("Can't write to location `%s`" % location)
        logger.debug("Writing `%s` sources to `%s`" % (self.handle, file_name))
        with open(file_name, "w") as f:
            header_keywords = []
            if self.is_default:
                header_keywords.append("default")
            header_keywords += ["handle", self.handle]
            f.write("#%s\n" % ":".join(header_keywords))
            csv_writer = csv.DictWriter(f, self.rows[0].keys())
            csv_writer.writeheader()
            csv_writer.writerows(self.rows)
        return file_name

    def from_set(self, src_set):
        """
        Use set to fill this Sources object. The set is expected to contain
        :param src_set: set with (rank, host) pairs
        :return: None
        """
        self.rows = [{"rank": str(rank), "hostname": hostname} for rank, hostname in src_set]

    def as_set(self, start=0, end=None):
        """
        Return rows of this sources list as a set. The set does not pertain any of
        the sources' meta data (DB handle, default). You can specify `start` and `end`
        to select just a chunk of data from the rows.

        Warning: There is no plausibility checking on `start` and `end` parameters.

        :param start: optional int marking beginning of chunk
        :param end: optional int marking end of chunk
        :return: set of (int rank, str hostname) pairs
        """
        if len(self.rows) == 0:
            return set()
        if end is None:
            end = len(self.rows)
        if "rank" in self.rows[0].keys():
            return set([(int(row["rank"]), row["hostname"]) for row in self.rows[start:end]])
        else:
            return set([(0, row["hostname"]) for row in self.rows[start:end]])

    def iter_chunks(self, chunk_size=None, min_chunk_size=1, chunk_start=0, chunk_stop=None):
        """
        Start iterating rows in chunks. chunk_size defaults to 1/20 of overall iteration size,
        but not to less than min_chunk_size. Returns generator function of Sources.next_chunk
        to call for obtaining the next chunk.

        This method is not re-entrant. Do not try to iterate the same Sources object concurrently.

        :param chunk_size: int
        :param min_chunk_size: int
        :param chunk_start: int
        :param chunk_stop: int
        :return: function
        """
        self.chunk_start = chunk_start
        self.chunk_stop = len(self) if chunk_stop is None else min(chunk_stop, len(self))
        self.chunk_size = (self.chunk_stop - self.chunk_start) / 20 if chunk_size is None else chunk_size
        self.chunk_size = max(self.chunk_size, min_chunk_size)
        self.chunk_offset = self.chunk_start
        return self.next_chunk

    def next_chunk(self, chunk_size=None, as_set=False):
        """
        Generator for obtaining next chunk as configured by .iter_chunk(). You can change chunk size on
        the fly by specifying chunk_size. The new size is remembered, but not subject to min_chunk_size
        given to .inter_chunks(). Returns None if chunk_stop was reached

        This method is not re-entrant. Do not try to iterate the same Sources object concurrently.

        :param chunk_size: int
        :param as_set: bool
        :return: list of rows or set of (int rank, str hostname) or None
        """
        # Iteration might have completed
        if self.chunk_offset >= self.chunk_stop:
            return None
        # Set new chunk size if given
        if chunk_size is not None:
            self.chunk_size = chunk_size
        # There might not be enough to fill a whole chunk
        current_chunk_start = self.chunk_offset
        current_chunk_stop = min(current_chunk_start + self.chunk_size, self.chunk_stop)
        self.chunk_offset = current_chunk_stop
        if as_set:
            return self.as_set(current_chunk_start, current_chunk_stop)
        else:
            return self.rows[current_chunk_start:current_chunk_stop]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

from distutils.spawn import find_executable
import logging
import os
import subprocess
import sys

import cache


logger = logging.getLogger(__name__)


def get_list(onecrl_env, workdir, commit, use_cache=True, cache_timeout=60*60):
    global logger

    dc = cache.DiskCache(os.path.join(workdir, "cache"), cache_timeout, purge=True)
    cache_id = "%s_revocations.txt" % onecrl_env
    if not use_cache:
        # Enforce re-extraction even if cached
        dc.delete(cache_id)
    cache_file = dc[cache_id]

    if cache_id not in dc:

        # Find Go binary
        go_bin = find_executable("go")
        if go_bin is None:
            logger.critical("Cannot find Go compiler")
            sys.exit(5)
        logger.debug("Using Go compiler at `%s`" % go_bin)

        # Prepare Go environment within our workdir
        go_path = os.path.join(workdir, "go")
        logger.debug("Using GOPATH `%s`" % go_path)
        go_env = os.environ.copy()
        go_env["GOPATH"] = go_path

        # Install / update oneCRL2RevocationsTxt package
        package = "github.com/mozilla/OneCRL-Tools/oneCRL2RevocationsTxt"
        repo_dir = os.path.join(go_path, "src", "github.com", "mozilla", "OneCRL-Tools")
        # If the package has already been downloaded, checkout master, else `go get` will fail
        if os.path.isdir(os.path.join(repo_dir, ".git")):
            logger.debug("Checking out commit `master` in `%s` for update" % repo_dir)
            if subprocess.call(["git", "checkout", "-q", "master"], cwd=repo_dir) != 0:
                logger.critical("Cannot checkout OneCRL-Tools git commit `master`")
                sys.exit(5)
        # `go get` internally uses `git pull`. `-d` prevents installation
        logger.debug("Installing / updating Go package `%s`" % package)
        if subprocess.call([go_bin, "get", "-u", "-d", package], env=go_env) != 0:
            logger.critical("Cannot get Go package `%s`" % package)
            sys.exit(5)
        # Checkout a known-working commit before installation
        logger.debug("Checking out commit `%s` in `%s`" % (commit, repo_dir))
        if subprocess.call(["git", "checkout", "-q", commit], cwd=repo_dir) != 0:
            logger.critical("Cannot checkout OneCRL-Tools git commit `%s`" % commit)
            sys.exit(5)
        # Install package
        if subprocess.call([go_bin, "install", package], env=go_env) != 0:
            logger.critical("Cannot install Go package `%s`" % package)
            sys.exit(5)

        # Run OneCRL Go binary to retrieve OnceCRL data
        if sys.platform == "win32":
            onecrl_bin = os.path.join(go_path, "bin", "oneCRL2RevocationsTxt.exe")
        else:
            onecrl_bin = os.path.join(go_path, "bin", "oneCRL2RevocationsTxt")
        if not os.path.isfile(onecrl_bin):
            logger.critical("Go package `oneCRL2RevocationsTxt` is missing executable")
            sys.exit(5)
        onecrl_cmd = [onecrl_bin, "--onecrlenv", onecrl_env]
        logger.debug("Running shell command `%s`" % " ".join(onecrl_cmd))
        try:
            revocations_data = subprocess.check_output(onecrl_cmd, env=go_env)
        except subprocess.CalledProcessError as error:
            logger.critical("Could not fetch revocations data: %s" % error)
            sys.exit(5)

        # oneCRL2RevocationsTxt does not indicate failure, but the result is empty.
        # Can we be sure this can never happen during regular operation?
        # See https://github.com/mozmark/OneCRL-Tools/issues/3
        if len(revocations_data) == 0:
            logger.critical("Revocations data was empty. Likely network failure.")
            sys.exit(5)

        # Write OneCRL data to cache
        logger.debug("Caching OneCRL revocations data in `%s`" % cache_file)
        with open(cache_file, "w") as f:
            f.write(revocations_data)

    else:
        logger.warning("Using cached OneCRL revocations data from `%s`" % cache_file)

    return cache_file



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import logging
import os
from shutil import rmtree
from time import time


logger = logging.getLogger(__name__)


class DiskCache(object):

    def __init__(self, root_directory, maximum_age=24*60*60, purge=False):
        global logger
        self.__root = os.path.abspath(root_directory)
        self.__maximum_age = maximum_age
        if not os.path.exists(self.__root):
            logger.debug('Creating cache directory %s' % self.__root)
            os.makedirs(self.__root)
        if purge:
            self.purge()

    def list(self):
        """Return list of IDs in cache"""
        return os.listdir(self.__root)

    def __clear(self):
        """Remove all entries from cache directory"""
        global logger
        logger.debug("Clearing cache directory `%s`" % self.__root)
        for cache_id in self:
            path = self[cache_id]
            if os.path.isdir(path):
                rmtree(path)
            else:
                os.remove(path)

    def delete(self, id_or_path=None):
        """
        Delete cache entries by ID or full path.
        Delete everything if no ID or path given.
        """
        if id_or_path is None:
            self.__clear()
            return
        path = self[id_or_path.lstrip(self.__root)]
        if os.path.isdir(path):
            rmtree(path)
        elif os.path.exists(path):
            os.remove(path)

    def purge(self, maximum_age=None):
        """Remove stale entries from cache directory"""
        global logger
        logger.debug("Purging stale cache entries from `%s`" % self.__root)
        if maximum_age is None:
            maximum_age = self.__maximum_age
        now = time()  # Current time as epoch
        stale_limit = now - maximum_age
        for cache_id in self:
            path = self[cache_id]
            mtime = os.path.getmtime(path)  # Modification time as epoch (might have just 1.0s resolution)
            if mtime < stale_limit:
                logger.debug('Purging stale cache entry `%s`' % cache_id)
                if os.path.isdir(path):
                    rmtree(path)
                else:
                    os.remove(path)

    def __iter__(self):
        """Iterate IDs of cache entries"""
        for name in self.list():
            yield name

    def __contains__(self, cache_id):
        """Check for presence of ID in cache"""
        return cache_id in self.list()

    def __getitem__(self, cache_id):
        """Return full path of cache entry (existing or not)"""
        return os.path.join(self.__root, cache_id)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import argparse
import logging
import coloredlogs
import os
import pkg_resources
import shutil
import sys
import tempfile
import threading
import time

import cleanup
import firefox_downloader as fd
import loader
import modes
import sources_db as sdb


# Initialize coloredlogs
logging.Formatter.converter = time.gmtime
logger = logging.getLogger(__name__)
coloredlogs.DEFAULT_LOG_FORMAT = "%(asctime)s %(levelname)s %(threadName)s %(name)s %(message)s"
coloredlogs.install(level="INFO")


def parse_args(argv=None):
    """
    Argument parsing. Parses from sys.argv if argv is None.
    :param argv: argument vector to parse
    :return: parsed arguments
    """
    if argv is None:
        argv = sys.argv[1:]

    pkg_version = pkg_resources.require("tlscanary")[0].version
    home = os.path.expanduser("~")

    # Set up the parent parser with shared arguments
    parser = argparse.ArgumentParser(prog="tlscanary")
    parser.add_argument("--version", action="version", version="%(prog)s " + pkg_version)
    parser.add_argument("-d", "--debug",
                        help="Enable debug",
                        action="store_true")
    parser.add_argument("-w", "--workdir",
                        help="Path to working directory",
                        type=os.path.abspath,
                        action="store",
                        default="%s/.tlscanary" % home)

    # Set up subparsers, one for each mode
    subparsers = parser.add_subparsers(help="Run mode", dest="mode")
    for mode_name in modes.all_modes:
        mode_class = modes.all_modes[mode_name]
        sub_parser = subparsers.add_parser(mode_name, help=mode_class.help)
        mode_class.setup_args(sub_parser)

    return parser.parse_args(argv)


tmp_dir = None
module_dir = None


def __create_tempdir():
    """
    Helper function for creating the temporary directory.
    Writes to the global variable tmp_dir
    :return: Path of temporary directory
    """
    temp_dir = tempfile.mkdtemp(prefix='tlscanary_')
    logger.debug('Created temp dir `%s`' % temp_dir)
    return temp_dir


class RemoveTempDir(cleanup.CleanUp):
    """
    Class definition for cleanup helper responsible
    for deleting the temporary directory prior to exit.
    """
    @staticmethod
    def at_exit():
        global tmp_dir
        if tmp_dir is not None:
            logger.debug('Removing temp dir `%s`' % tmp_dir)
            shutil.rmtree(tmp_dir, ignore_errors=True)


restore_terminal_encoding = None


def get_terminal_encoding():
    """
    Helper function to get current terminal encoding
    """
    global logger
    if sys.platform.startswith("win"):
        logger.debug("Running `chcp` shell command")
        chcp_output = os.popen("chcp").read().strip()
        logger.debug("chcp output: `%s`" % chcp_output)
        if chcp_output.startswith("Active code page:"):
            codepage = chcp_output.split(": ")[1]
            logger.debug("Active codepage is `%s`" % codepage)
            return codepage
        else:
            logger.warning("There was an error detecting the active codepage")
            return None
    else:
        logger.debug("Platform does not require switching terminal encoding")
        return None


def set_terminal_encoding(encoding):
    """
    Helper function to set terminal encoding.
    """
    global logger
    if sys.platform.startswith("win"):
        logger.debug("Running `chcp` shell command, setting codepage to `%s`", encoding)
        chcp_output = os.popen("chcp %s" % encoding).read().strip()
        logger.debug("chcp output: `%s`" % chcp_output)
        if chcp_output == "Active code page: %s" % encoding:
            logger.debug("Successfully set codepage to `%s`" % encoding)
        else:
            logger.warning("Can't set codepage for terminal")


def fix_terminal_encoding():
    """
    Helper function to set terminal to platform-specific UTF encoding
    """
    global restore_terminal_encoding
    restore_terminal_encoding = get_terminal_encoding()
    if restore_terminal_encoding is None:
        return
    if sys.platform.startswith("win"):
        platform_utf_encoding = "65001"
    else:
        platform_utf_encoding = None
    if restore_terminal_encoding != platform_utf_encoding:
        set_terminal_encoding(platform_utf_encoding)


class ResetTerminalEncoding(cleanup.CleanUp):
    """
    Class for restoring original terminal encoding at exit.
    """
    @staticmethod
    def at_exit():
        global restore_terminal_encoding
        if restore_terminal_encoding is not None:
            set_terminal_encoding(restore_terminal_encoding)


# This is the entry point used in setup.py
def main(argv=None):
    global logger, tmp_dir, module_dir

    module_dir = os.path.split(__file__)[0]

    args = parse_args(argv)

    if args.debug:
        coloredlogs.install(level='DEBUG')

    logger.debug("Command arguments: %s" % args)

    cleanup.init()
    fix_terminal_encoding()
    tmp_dir = __create_tempdir()

    # If 'list' is specified as test, list available test sets, builds, and platforms
    if "source" in args and args.source == "list":
        coloredlogs.install(level='ERROR')
        db = sdb.SourcesDB(args)
        build_list, platform_list, _, _ = fd.FirefoxDownloader.list()
        print "Available builds: %s" % ' '.join(build_list)
        print "Available platforms: %s" % ' '.join(platform_list)
        print "Available test sets:"
        for handle in db.list():
            test_set = db.read(handle)
            if handle == db.default:
                default = " (default)"
            else:
                default = ""
            print "  - %s [%d hosts]%s" % (handle, len(test_set), default)
        return 0

    # Create workdir (usually ~/.tlscanary, used for caching etc.)
    # Assumes that no previous code must write to it.
    if not os.path.exists(args.workdir):
        logger.debug('Creating working directory %s' % args.workdir)
        os.makedirs(args.workdir)

    # Load the specified test mode
    try:
        loader.run(args, module_dir, tmp_dir)

    except KeyboardInterrupt:
        logger.critical("\nUser interrupt. Quitting...")
        return 10

    if len(threading.enumerate()) > 1:
        logger.info("Waiting for background threads to finish")
        while len(threading.enumerate()) > 1:
            logger.debug("Remaining threads: %s" % threading.enumerate())
            time.sleep(2)

    return 0



#!/usr/bin/env python2

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

from cryptography import x509
from cryptography.hazmat import backends
from cryptography.hazmat.primitives import serialization
from cryptography.x509.oid import ExtensionOID


class Cert(object):
    """Class for handling X509 certificates"""

    def __init__(self, data):
        """
        Cert constructor

        It can handle PEM and DER encoded strings and lists of int bytes.

        :param data: str or list of int
        """
        if type(data) == list:
            data = ''.join([chr(x) for x in data])
        if type(data) != str:
            raise Exception("data must be string or list of int bytes")
        self.__raw_data = data
        if "-----BEGIN CERTIFICATE-----" in data:
            self.x509 = x509.load_pem_x509_certificate(data, backends.default_backend())
            self.__raw_type = "PEM"
        else:
            self.x509 = x509.load_der_x509_certificate(data, backends.default_backend())
            self.__raw_type = "DER"

    def as_pem(self):
        """
        Convert certificate to PEM-encoded string

        :return: str
        """
        return self.x509.public_bytes(encoding=serialization.Encoding.PEM)

    def as_der(self):
        """
        Convert certificate to DER-encoded string

        :return: str
        """
        return self.x509.public_bytes(encoding=serialization.Encoding.DER)

    def signature_hash_algorithm(self):
        """
        Extract certificate's hash algorithm

        :return: str
        """
        return self.x509.signature_hash_algorithm.name

    def subject_alt_name(self):
        """
        Extract certificate's alt names

        :return: unicode
        """
        try:
            alt_names = self.x509.extensions.get_extension_for_oid(ExtensionOID.SUBJECT_ALTERNATIVE_NAME).value
            alt_name_strings = [alt_name.value for alt_name in alt_names]
            return ",".join(alt_name_strings)
        except x509.ExtensionNotFound:
            return "(no subject alt name)"

    def ext_key_usage(self):
        """
        Extract certificate's permitted extended usages

        :return: str
        """
        try:
            usages = self.x509.extensions.get_extension_for_oid(ExtensionOID.EXTENDED_KEY_USAGE).value
            usages_strings = [usage._name for usage in usages]
            return ",".join(usages_strings)
        except x509.ExtensionNotFound:
            return "(no ext key usage)"



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import datetime
import logging
import pkg_resources as pkgr
import sys

from regression import RegressionMode
import tlscanary.runlog as rl


logger = logging.getLogger(__name__)


class PerformanceMode(RegressionMode):

    name = "performance"
    help = "Run a performance regression test on two Firefox versions"

    def __init__(self, args, module_dir, tmp_dir):
        global logger

        super(PerformanceMode, self).__init__(args, module_dir, tmp_dir)

        # Define instance attributes for later use
        self.start_time = None
        self.test_uri_set = None
        self.base_uri_set = None
        self.total_change = None

    def setup(self):
        # Additional argument validation for hard-coded limits,
        # because 1000 URIs x 20 scans per URI x 2 builds is a lot of data
        # will investigate upper limit later
        if self.args.limit > 1000:
            logger.warning("Limiting performance tests to 1000 hosts.")
            self.args.limit = 1000
        if self.args.scans > 20:
            logger.critical("Limiting performance test to 20 scans per URI list for now")
            sys.exit(1)
        super(PerformanceMode, self).setup()

    def run(self):
        # Perform the scan
        self.start_time = datetime.datetime.now()

        meta = {
            "tlscanary_version": pkgr.require("tlscanary")[0].version,
            "mode": self.name,
            "args": vars(self.args),
            "argv": sys.argv,
            "sources_size": len(self.sources),
            "test_metadata": self.test_metadata,
            "base_metadata": self.base_metadata,
            "run_start_time": datetime.datetime.utcnow().isoformat()
        }

        rldb = rl.RunLogDB(self.args)
        log = rldb.new_log()
        log.start(meta=meta)

        test_uri_sets = []
        base_uri_sets = []

        self.total_change = 0
        test_speed_aggregate = 0
        base_speed_aggregate = 0

        for i in xrange(0, self.args.scans):
            test_uri_sets.append(self.run_test(self.test_app, self.sources, profile=self.test_profile,
                                               prefs=self.args.prefs_test, get_info=True,
                                               get_certs=not self.args.remove_certs, return_only_errors=False))

            base_uri_sets.append(self.run_test(self.base_app, self.sources, profile=self.base_profile,
                                               prefs=self.args.prefs_base, get_info=True,
                                               get_certs=not self.args.remove_certs, return_only_errors=False))

        # extract connection speed from all scans
        test_connections_all = []
        for uri_set in test_uri_sets:
            test_connections_all.append(self.extract_connection_speed(uri_set))

        base_connections_all = []
        for uri_set in base_uri_sets:
            base_connections_all.append(self.extract_connection_speed(uri_set))

        # collapse all scan data into one URI set
        self.consolidate_connection_speed_info(test_uri_sets[0], test_connections_all)
        self.consolidate_connection_speed_info(base_uri_sets[0], base_connections_all)

        # the first URI set becomes our primary set
        self.test_uri_set = test_uri_sets[0]
        self.base_uri_set = base_uri_sets[0]

        # new values to be inserted into response
        for test_record in self.test_uri_set:
            base_record = [d for d in self.base_uri_set if d[1] == test_record[1]][0]
            test_response_time = float(test_record[2].response.connection_speed_average)
            base_response_time = float(base_record[2].response.connection_speed_average)
            test_speed_aggregate += test_response_time
            base_speed_aggregate += base_response_time
            pct_change = float((test_response_time - base_response_time) / base_response_time) * 100
            test_record[2].response.connection_speed_change = int(pct_change)
            # save the speed samples of the base record to the test record for now,
            # in case we decide we want to include this in the report later
            test_record[2].response.connection_speed_base_samples = base_record[2].response.connection_speed_samples

        self.total_change = float((test_speed_aggregate - base_speed_aggregate) / base_speed_aggregate) * 100

        for rank, host, result in self.test_uri_set:
            log.log(result.as_dict())

        meta["run_finish_time"] = datetime.datetime.utcnow().isoformat()
        meta["total_change"] = self.total_change
        self.save_profile(self.test_profile, "test_profile", log)
        self.save_profile(self.base_profile, "base_profile", log)
        log.stop(meta=meta)

    @staticmethod
    def extract_connection_speed(uri_set):
        new_set = []
        for record in uri_set:
            new_set.append([
                record[0],  # URI rank
                record[2].response.response_time - record[2].response.command_time  # connection speed
            ])
        return new_set

    def consolidate_connection_speed_info(self, uri_set, connection_sets):
        for record in uri_set:
            temp_speeds = []
            speed_aggregate = 0
            for connection_set in connection_sets:
                speed = [d for d in connection_set if d[0] == record[0]][0][1]
                speed_aggregate += speed
                temp_speeds.append(speed)
            record[2].response.connection_speed_average = speed_aggregate / self.args.scans
            record[2].response.connection_speed_samples = str(temp_speeds).strip('[]')



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import datetime
import logging
import pkg_resources as pkgr
import sys

from basemode import BaseMode
import tlscanary.progress as pr
import tlscanary.runlog as rl
import tlscanary.sources_db as sdb

logger = logging.getLogger(__name__)


class RegressionMode(BaseMode):
    name = "regression"
    help = "Run a TLS regression test on two Firefox versions"

    def __init__(self, args, module_dir, tmp_dir):
        global logger

        super(RegressionMode, self).__init__(args, module_dir, tmp_dir)

        # Define instance attributes for later use
        self.start_time = None
        self.test_app = None
        self.base_app = None
        self.test_metadata = None
        self.base_metadata = None
        self.test_profile = None
        self.base_profile = None
        self.altered_profile = None
        self.sources = None
        self.revoked_source = None
        self.custom_ocsp_pref = None

    def one_crl_sanity_check(self):
        global logger

        # Query host(s) with a known revoked cert and examine the results
        # These hosts must be revoked via OCSP and/or OneCRL
        db = sdb.SourcesDB(self.args)
        self.revoked_source = db.read("revoked")
        logger.debug("%d host(s) in revoked test set" % len(self.revoked_source))
        next_chunk = self.revoked_source.iter_chunks(chunk_size=1/50, min_chunk_size=1000)
        host_set_chunk = next_chunk(as_set=True)

        # Note: turn off OCSP for this test, to factor out that mechanism
        self.custom_ocsp_pref = ["security.OCSP.enabled;0"]

        # First, use the test build and profile as-is
        # This should return errors, which means OneCRL is working
        test_result = self.run_test(self.test_app, url_list=host_set_chunk, profile=self.test_profile,
                                    prefs=self.custom_ocsp_pref, num_workers=1, n_per_worker=1)

        # Second, use the test build with a profile that is missing OneCRL entries
        # This should NOT return errors, which means we've turned off protection
        self.altered_profile = self.make_profile("altered_profile", "none")

        base_result = self.run_test(self.test_app, url_list=host_set_chunk, profile=self.altered_profile,
                                    prefs=self.custom_ocsp_pref, num_workers=1, n_per_worker=1)

        logger.debug("Length of first OneCRL check, with revocation: %d" % len(test_result))
        logger.debug("Length of second OneCRL check, without revocation: %d" % len(base_result))

        # If our list of revoked sites are all blocked, and we can verify
        # that they can be unblocked, this confirms that OneCRL is working
        if len(test_result) == len(self.revoked_source) and len(base_result) == 0:
            return True
        else:
            return False

    def setup(self):
        global logger

        # Argument validation logic
        # Make sure user has test build
        if self.args.test is None:
            logger.critical("Must specify test build for regression testing")
            sys.exit(5)
        elif self.args.base is None:
            logger.critical("Must specify base build for regression testing")
            sys.exit(5)

        if self.args.scans < 2:
            logger.critical("Must specify minimum of 2 scans for regression testing")
            sys.exit(5)

        if self.args.prefs is not None:
            if self.args.prefs_test is not None or self.args.prefs_base is not None:
                logger.warning("Detected both global prefs and individual build prefs.")

        self.test_app = self.get_test_candidate(self.args.test)
        self.base_app = self.get_test_candidate(self.args.base)

        self.test_metadata = self.collect_worker_info(self.test_app)
        self.base_metadata = self.collect_worker_info(self.base_app)

        # Setup custom profiles
        self.test_profile = self.make_profile("test_profile", self.args.onecrl)
        self.base_profile = self.make_profile("base_profile", "production")

        # Compile the set of hosts to test
        db = sdb.SourcesDB(self.args)
        logger.info("Reading `%s` host database" % self.args.source)
        self.sources = db.read(self.args.source)
        logger.info("%d hosts in test set" % len(self.sources))

        # Sanity check for OneCRL - if it fails, abort run
        if not self.one_crl_sanity_check():
            logger.critical("OneCRL sanity check failed, aborting run")
            sys.exit(5)

    def run(self):
        global logger

        logger.info("Testing Firefox %s %s against Firefox %s %s" %
                    (self.test_metadata["app_version"], self.test_metadata["branch"],
                     self.base_metadata["app_version"], self.base_metadata["branch"]))

        self.start_time = datetime.datetime.now()

        meta = {
            "tlscanary_version": pkgr.require("tlscanary")[0].version,
            "mode": self.name,
            "args": vars(self.args),
            "argv": sys.argv,
            "sources_size": len(self.sources),
            "test_metadata": self.test_metadata,
            "base_metadata": self.base_metadata,
            "run_start_time": datetime.datetime.utcnow().isoformat()
        }

        rldb = rl.RunLogDB(self.args)
        log = rldb.new_log()
        log.start(meta=meta)
        progress = pr.ProgressTracker(total=len(self.sources), unit="hosts", average=30*60.0)

        limit = len(self.sources) if self.args.limit is None else self.args.limit

        # Split work into 50 chunks to conserve memory, but make no chunk smaller than 1000 hosts
        next_chunk = self.sources.iter_chunks(chunk_size=limit/50, min_chunk_size=1000)

        try:
            while True:
                host_set_chunk = next_chunk(as_set=True)
                if host_set_chunk is None:
                    break

                logger.info("Starting regression run on chunk of %d hosts" % len(host_set_chunk))

                error_set = self.run_regression_passes(host_set_chunk, report_completed=progress.log_completed,
                                                       report_overhead=progress.log_overhead)
                # Log progress per chunk
                logger.info("Progress: %s" % str(progress))

                # Commit results to log
                for rank, host, result in error_set:
                    log.log(result.as_dict())

        except KeyboardInterrupt:
            logger.critical("Ctrl-C received")
            progress.stop_reporting()
            raise KeyboardInterrupt

        finally:
            progress.stop_reporting()

        meta["run_finish_time"] = datetime.datetime.utcnow().isoformat()
        self.save_profile(self.test_profile, "test_profile", log)
        self.save_profile(self.base_profile, "base_profile", log)
        self.save_profile(self.altered_profile, "altered_profile", log)
        log.stop(meta=meta)

    def run_regression_passes(self, host_set, report_completed=None, report_overhead=None):
        global logger
        # Compile set of error hosts in multiple scans

        # Each scan:
        # - Runs full test set against the test candidate
        # - Runs new error set against baseline candidate
        # - Take any remaining errors and repeat the above steps

        current_host_set = host_set
        num_workers = self.args.parallel
        requests_per_worker = self.args.requestsperworker
        timeout = self.args.timeout
        max_timeout = self.args.max_timeout

        for current_scan in xrange(1, self.args.scans + 1):

            # Specify different callback only for initial test scan
            if current_scan == 1:
                report_callback_value = report_completed
            else:
                report_callback_value = report_overhead

            # Actual test running for both builds
            test_error_set = self.run_test(self.test_app, current_host_set, profile=self.test_profile,
                                           prefs=self.args.prefs_test, num_workers=num_workers,
                                           n_per_worker=requests_per_worker, timeout=timeout,
                                           report_callback=report_callback_value)
            logger.info("Scan #%d with test candidate yielded %d error hosts"
                        % (current_scan, len(test_error_set)))
            logger.debug("Scan #%d test candidate errors: %s"
                         % (current_scan, ' '.join(["%d,%s" % (r, u) for r, u in test_error_set])))
            base_error_set = self.run_test(self.base_app, test_error_set, profile=self.base_profile,
                                           prefs=self.args.prefs_base, num_workers=num_workers,
                                           n_per_worker=requests_per_worker, timeout=timeout,
                                           report_callback=report_overhead)
            logger.info("Scan #%d with baseline candidate yielded %d error hosts"
                        % (current_scan, len(base_error_set)))
            logger.debug("Scan #%d baseline candidate errors: %s"
                         % (current_scan, ' '.join(["%d,%s" % (r, u) for r, u in base_error_set])))

            current_host_set = test_error_set.difference(base_error_set)

            # If no errors, no need to keep scanning
            if len(current_host_set) == 0:
                break
            else:
                # Slow down number of workers and scans with each pass
                # to make results more precise
                num_workers = max(1, int(num_workers * 0.75))
                requests_per_worker = max(1, int(requests_per_worker * 0.75))
                timeout = min(max_timeout, timeout * 1.25)

        last_error_set = current_host_set

        # Final pass, information extraction only:
        # - Run error set from previous pass against the test candidate with minimal workers, requests
        # - Have workers return extra runtime information, including certificates

        logger.debug("Extracting runtime information from %d hosts" % (len(last_error_set)))
        final_error_set = self.run_test(self.test_app, last_error_set, profile=self.test_profile,
                                        prefs=self.args.prefs_test, num_workers=1, n_per_worker=1,
                                        get_info=True, get_certs=not self.args.remove_certs,
                                        report_callback=report_overhead)

        if len(final_error_set) > 0:
            logger.warning("%d potential regressions found: %s"
                           % (len(final_error_set), ' '.join(["%d,%s" % (r, u) for r, u, d in final_error_set])))

        # Find out if the information extraction pass changed the results
        if self.args.debug:
            # Final set includes additional result data, so filter that out before comparison
            stripped_final_set = set()

            for rank, host, data in final_error_set:
                stripped_final_set.add((rank, host))

            if stripped_final_set != last_error_set:
                diff_set = last_error_set.difference(stripped_final_set)
                logger.debug("Number of hosts dropped out of final error set: %d" % len(diff_set))
                logger.debug(diff_set)

        return final_error_set



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import json
import logging
import os
import sys

from basemode import BaseMode
import tlscanary.report as report
import tlscanary.runlog as rl


logger = logging.getLogger(__name__)


class LogMode(BaseMode):
    """
    Mode to access run logs and generate reports
    """

    name = "log"
    help = "Query and maintain the run log database and create reports"
    logging_modes = ["performance", "regression", "scan"]

    @classmethod
    def setup_args(cls, parser):

        group = parser.add_argument_group("log actions")

        group.add_argument("-a", "--action",
                           help="Action to perform (default: list)",
                           choices=["delete", "webreport", "json", "list"],
                           action="store",
                           default="list")

        group.add_argument("--really",
                           help="Sanity flag required by dangerous actions like `delete`",
                           action="store_true")

        group = parser.add_argument_group("log selection")
        # Breaking alphabetical ordering, because logically `include` happens before `exclude`
        group.add_argument("-i", "--include",
                           help=("Include certain logs in log actions. Can be any of "
                                 "{all,complete,incomplete,incompatible}, any run mode of {%s}, an integer number "
                                 "specifying the N latest logs, or any log handle. "
                                 "Can be specified repeatedly. (default: all)") % ",".join(cls.logging_modes),
                           action="append",
                           default=[])

        group.add_argument("-e", "--exclude",
                           help=("Exclude certain logs from log actions. Can be any of "
                                 "{complete,incomplete,incompatble}, any run mode of {%s}, an integer number "
                                 "specifying the N latest logs, or any log handle. "
                                 "Can be specified repeatedly. (default: None") % ",".join(cls.logging_modes),
                           action="append",
                           default=[])

        group = parser.add_argument_group("report generation")
        group.add_argument("-o", "--output",
                           help="Path to output directory for writing reports",
                           type=os.path.abspath,
                           action="store",
                           default=None)

    def __init__(self, args, module_dir, tmp_dir):
        super(LogMode, self).__init__(args, module_dir, tmp_dir)
        self.log_db = rl.RunLogDB(self.args)

        # Check arguments
        if len(self.args.include) == 0:
            logger.debug("No --include argument given, defaulting to `all`")
            self.args.include.append("all")

    def run(self):
        global logger

        log_handles = self.log_db.list()

        # Compile dict of all logs as {log handle => log object}
        all_logs = dict(zip(log_handles, map(self.log_db.read_log, log_handles)))

        # Filter log list according to includes and excludes
        log_list = self.compile_match_list(all_logs, self.args.include, self.args.exclude)
        if log_list is None:
            sys.exit(5)

        # See what to do with those logs
        if self.args.action is None or self.args.action == "list":
            self.print_log_list(log_list)

        elif self.args.action == "delete":
            if not self.args.really:
                for log_name in sorted(log_list.keys()):
                    logger.info("Would delete log `%s`" % log_name)
                logger.critical("Is this what you --really want?")
                sys.exit(0)
            else:
                for log_name in sorted(log_list.keys()):
                    logger.info("Deleting log `%s`" % log_name)
                    log_list[log_name].delete()

        elif self.args.action == "json":
            log_data = []
            for log_name in log_list:
                log = self.log_db.read_log(log_name)
                meta = log.get_meta()
                if not meta["run_completed"]:
                    logger.warning("Skipping incomplete log `%s`" % log_name)
                    continue
                log_data.append({"meta": log.get_meta(), "data": [line for line in log]})

            print json.dumps(log_data, indent=4, sort_keys=True)

        elif self.args.action == "jsonreport":
            if self.args.output is None:
                logger.critical("You must specify -o/--output for writing the JSON report")
                sys.exit(5)
            report.generate("json", log_list, self.args.output)

        elif self.args.action == "webreport":
            if self.args.output is None:
                logger.critical("You must specify -o/--output for writing the HTML report")
                sys.exit(5)
            report.generate("web", log_list, self.args.output)

        else:
            logger.critical("Report action `%s` not implemented" % self.args.action)
            sys.exit(5)

    def compile_match_list(self, all_logs, include, exclude):
        global logger

        # TODO: method needs testing
        # Compile list of included logs
        matching_logs = {}
        for arg in include:
            if arg == "all":
                matching_logs.update(all_logs)
            elif arg == "complete":
                for log_name in all_logs:
                    log = all_logs[log_name]
                    if log.has_finished():
                        matching_logs[log_name] = log
            elif arg == "incomplete":
                for log_name in all_logs:
                    log = all_logs[log_name]
                    if not log.has_finished():
                        matching_logs[log_name] = log
            elif arg == "incompatible":
                for log_name in all_logs:
                    log = all_logs[log_name]
                    if not log.is_compatible():
                        matching_logs[log_name] = log
            elif arg in self.logging_modes:
                for log_name in all_logs:
                    log = all_logs[log_name]
                    if log.get_meta()["mode"] == arg:
                        matching_logs[log_name] = log
            elif arg in all_logs:
                matching_logs[arg] = all_logs[arg]
            elif arg.isdigit():
                last_n = int(arg)
                for log_name in sorted(all_logs.keys())[-last_n:]:
                    log = all_logs[log_name]
                    matching_logs[log_name] = log
            else:
                logger.critical("Invalid -i/--include argument: `%s`" % arg)
                return None

        logger.debug("Included logs: %s" % sorted(matching_logs.keys()))

        # Throw away excluded logs
        for arg in exclude:
            if arg == "complete":
                for log_name in all_logs:
                    log = all_logs[log_name]
                    if log.has_finished() and log_name in matching_logs:
                        del(matching_logs[log_name])
            elif arg == "incomplete":
                for log_name in all_logs:
                    log = all_logs[log_name]
                    if not log.has_finished() and log_name in matching_logs:
                        del(matching_logs[log_name])
            elif arg == "incompatible":
                for log_name in all_logs:
                    log = all_logs[log_name]
                    if not log.is_compatible() and log_name in matching_logs:
                        del (matching_logs[log_name])
            elif arg in self.logging_modes:
                for log_name in all_logs:
                    log = all_logs[log_name]
                    if log.get_meta()["mode"] == arg and log_name in matching_logs:
                        del(matching_logs[log_name])
            elif arg in all_logs:
                if arg in matching_logs:
                    del(matching_logs[arg])
            elif arg.isdigit():
                last_n = int(arg)
                for log_name in sorted(all_logs.keys())[-last_n:]:
                    if log_name in matching_logs:
                        del (matching_logs[log_name])
            else:
                logger.critical("Invalid -e/--exclude argument: `%s`" % arg)
                return None

        logger.debug("Included logs after exclusion: %s" % sorted(matching_logs.keys()))
        return matching_logs

    @staticmethod
    def print_log_list(log_list):
        for log_name in sorted(log_list.keys()):
            log = log_list[log_name]
            meta = log.get_meta()
            mode = meta["mode"] if "mode" in meta else "unknown"
            incomplete_marker = "" if log.has_finished() else "(*)"
            if not log.is_compatible():
                print "%s\t\t\t\t\t\tINCOMPATIBLE LOG FORMAT" % log_name
            elif mode == "regression" or mode == "performance":
                size = os.path.getsize((log.part("log.bz2")))
                if size > 100*1024*1024:
                    logger.warning("Log `%s` contains %.1f MBytes of data. counting may take a while"
                                   % (log.handle, size/1024.0/10124.0))
                print "%s%s\tlines=%-6d\tmode=%-12s\tFx %s %s / %s vs. Fx %s %s / %s" % (
                    log_name,
                    incomplete_marker,
                    len(log),
                    mode,
                    meta["test_metadata"]["app_version"],
                    meta["test_metadata"]["branch"].capitalize(),
                    meta["test_metadata"]["nss_version"],
                    meta["base_metadata"]["app_version"],
                    meta["base_metadata"]["branch"].capitalize(),
                    meta["base_metadata"]["nss_version"])
            elif mode == "scan":
                size = os.path.getsize((log.part("log.bz2")))
                if size > 100*1024*1024:
                    logger.warning("Log `%s` contains %.1f MBytes of data. Counting may take a while"
                                   % (log.handle, size/1024.0/1024.0))
                print "%s%s\tlines=%-6d\tmode=%-12s\tFx %s %s / %s" % (
                    log_name,
                    incomplete_marker,
                    len(log),
                    mode,
                    meta["test_metadata"]["app_version"],
                    meta["test_metadata"]["branch"].capitalize(),
                    meta["test_metadata"]["nss_version"])
            else:
                print "%s%s\t lines=%-6d\tmode=%-12s" % (
                    log_name,
                    incomplete_marker,
                    meta["log_lines"],
                    mode)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import csv
import datetime
import logging
import os
import sys
import zipfile

from basemode import BaseMode
from tlscanary.firefox_downloader import get_to_file
import tlscanary.progress as pr
import tlscanary.sources_db as sdb


logger = logging.getLogger(__name__)


class SourceUpdateMode(BaseMode):
    """
    Mode to update host databases from publicly available top sites data
    """

    name = "srcupdate"
    help = "Update hosts databases used by other modes"

    # There are various top sites databases that might be considered for querying here.
    # The other notable database is the notorious `Alexa Top 1M` which is available at
    # "http://s3.amazonaws.com/alexa-static/top-1m.csv.zip". It is based on usage data
    # gathered from the equally notorious Alexa browser toolbar, while the `Umbrella top 1M`
    # used below is DNS-based and its ranking is hence considered to be more representative.
    # `Umbrella` and `Alexa` use precisely the same format and their links are thus
    # interchangeable.
    # For future reference, there is also Ulfr's database at
    # "https://ulfr.io/f/top1m_has_tls_sorted.csv". It requires a different parser but
    # has the advantage of clustering hosts by shared certificates.

    top_sites_location = "http://s3-us-west-1.amazonaws.com/umbrella-static/top-1m.csv.zip"

    def __init__(self, args, module_dir, tmp_dir):
        super(SourceUpdateMode, self).__init__(args, module_dir, tmp_dir)
        self.start_time = None
        self.db = None
        self.sources = None
        self.app = None
        self.profile = None

    def setup(self):
        global logger

        self.app = self.get_test_candidate(self.args.base)
        self.profile = self.make_profile("base_profile")

        tmp_zip_name = os.path.join(self.tmp_dir, "top.zip")
        logger.info("Fetching unfiltered top sites data from the `Umbrella Top 1M` online database")
        get_to_file(self.top_sites_location, tmp_zip_name)

        try:
            zipped = zipfile.ZipFile(tmp_zip_name)
            if len(zipped.filelist) != 1 or not zipped.filelist[0].orig_filename.lower().endswith(".csv"):
                logger.critical("Top sites zip file has unexpected content")
                sys.exit(5)
            tmp_csv_name = zipped.extract(zipped.filelist[0], self.tmp_dir)
        except zipfile.BadZipfile:
            logger.critical("Error opening top sites zip archive")
            sys.exit(5)

        self.db = sdb.SourcesDB(self.args)
        is_default = self.args.source == self.db.default
        self.sources = sdb.Sources(self.args.source, is_default)

        with open(tmp_csv_name) as f:
            cr = csv.DictReader(f, fieldnames=["rank", "hostname"])
            self.sources.rows = [row for row in cr]

        # A mild sanity check to see whether the downloaded data is valid.
        if len(self.sources) < 900000:
            logger.warning("Top sites is surprisingly small, just %d hosts" % len(self.sources))
        if self.sources.rows[0] != {"hostname": "google.com", "rank": "1"}:
            logger.warning("Top sites data looks weird. First line: `%s`" % self.sources.rows[0])

    def run(self):
        """
        Perform the filter run. The objective is to filter out permanent errors so
        we don't waste time on them during regular test runs.

        The concept is:
        Run top sites in chunks through Firefox and re-test all error URLs from that
        chunk a number of times to weed out spurious network errors. Stop the process
        once the required number of working hosts is collected.
        """
        global logger

        self.start_time = datetime.datetime.now()

        limit = 500000
        if self.args.limit is not None:
            limit = self.args.limit

        logger.info("There are %d hosts in the unfiltered host set" % len(self.sources))
        logger.info("Compiling set of %d working hosts for `%s` database update" % (limit, self.sources.handle))
        working_set = set()

        # Chop unfiltered sources data into chunks and iterate over each
        # .iter_chunks() returns a generator method to call for next chunk
        next_chunk = self.sources.iter_chunks(chunk_size=1000)
        chunk_size = self.sources.chunk_size

        progress = pr.ProgressTracker(total=limit, unit="hosts", average=60 * 60.0)

        try:
            while True:
                hosts_to_go = max(0, limit - len(working_set))
                # Check if we're done
                if hosts_to_go == 0:
                    break
                logger.info("%d hosts to go to complete the working set" % hosts_to_go)

                # Shrink chunk if it contains way more hosts than required to complete the working set
                #
                # CAVE: This assumes that this is the last chunk we require. The downsized chunk
                # is still 50% larger than required to complete the set to compensate for broken
                # hosts. If the error rate in the chunk is greater than 50%, another chunk will be
                # consumed, resulting in a gap of untested hosts between the end of this downsized
                # chunk and the beginning of the next. Not too bad, but important to be aware of.
                if chunk_size > hosts_to_go * 2:
                    chunk_size = min(chunk_size, hosts_to_go * 2)
                pass_chunk = next_chunk(chunk_size, as_set=True)

                # Check if we ran out of data for completing the set
                if pass_chunk is None:
                    logger.warning("Ran out of hosts to complete the working set")
                    break

                # Run chunk through multiple passes of Firefox, leaving only persistent errors in the
                # error set.
                pass_chunk_size = len(pass_chunk)
                chunk_end = self.sources.chunk_offset
                chunk_start = chunk_end - pass_chunk_size
                logger.info("Processing chunk of %d hosts from the unfiltered set (#%d to #%d)"
                            % (chunk_end - chunk_start, chunk_start, chunk_end - 1))
                pass_errors = pass_chunk

                for i in xrange(self.args.scans):

                    logger.info("Pass %d with %d hosts" % (i + 1, len(pass_errors)))

                    # First run is regular, every other run is overhead
                    if i == 0:
                        report_callback = None
                    else:
                        report_callback = progress.log_overhead

                    pass_errors = self.run_test(self.app, pass_errors, profile=self.profile, get_info=False,
                                                get_certs=False, return_only_errors=True,
                                                report_callback=report_callback)
                    len_pass_errors = len(pass_errors)

                    # Log progress of first pass
                    if i == 0:
                        progress.log_completed(pass_chunk_size - len_pass_errors)
                        progress.log_overhead(len_pass_errors)

                    if len_pass_errors == 0:
                        break

                logger.info("Error rate in chunk was %.1f%%"
                            % (100.0 * float(len_pass_errors) / float(chunk_end - chunk_start)))

                # Add all non-errors to the working set
                working_set.update(pass_chunk.difference(pass_errors))

                # Log progress after every chunk
                logger.info(str(progress))

        except KeyboardInterrupt:
            logger.critical("Ctrl-C received")
            raise KeyboardInterrupt

        final_src = sdb.Sources(self.sources.handle, is_default=self.sources.is_default)
        final_src.from_set(working_set)
        final_src.sort()
        final_src.trim(limit)

        if len(final_src) < limit:
            logger.warning("Ran out of hosts to complete the working set")

        logger.info("Collected %d working hosts for the updated test set" % len(final_src))
        logger.info("Writing updated `%s` host database" % final_src.handle)
        self.db.write(final_src)

    def teardown(self):
        # Free some memory
        self.db = None
        self.sources = None
        self.app = None
        self.profile = None



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

from distutils.spawn import find_executable
import logging
import os
import stat
import subprocess
import sys

import cache
from firefox_app import FirefoxApp


logger = logging.getLogger(__name__)


def extract(archive_file, workdir, cache_timeout=24*60*60, use_cache=True):
    """Extract a Firefox archive file into a subfolder in the given temp dir."""
    global logger

    logger.info("Extracting Firefox archive `%s`" % archive_file)

    # Find 7zip binary
    sz_bin = find_executable("7z")
    if sz_bin is None:
        logger.critical("Cannot find 7zip")
        sys.exit(5)
    logger.debug("Using 7zip executable at `%s`" % sz_bin)

    # Name in cache is file name without extensions
    cache_id = os.path.basename(archive_file)
    cache_id = os.path.splitext(cache_id)[0]
    cache_id = os.path.splitext(cache_id)[0]

    dc = cache.DiskCache(os.path.join(workdir, "cache"), cache_timeout, purge=True)

    if not use_cache:
        # Enforce re-extraction even if cached
        dc.delete(cache_id)

    cache_dir = dc[cache_id]

    if cache_id not in dc:
        cmd = [sz_bin, "x", "-y", "-bd", "-o%s" % cache_dir, archive_file]
        logger.debug("Executing shell command `%s`" % " ".join(cmd))
        try:
            output = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            logger.error("7zip failed: %s" % repr(e.output))
            raise Exception("Unable to extract Firefox archive")
        logger.debug("7zip succeeded: %s" % repr(output))

        # Check whether we have just extracted a tar file (from a .tar.bz2 archive)
        inner_tar = os.path.join(cache_dir, "%s.tar" % cache_id)
        if os.path.isfile(inner_tar):
            logger.debug("Running second 7zip pass on inner TAR archive `%s`" % inner_tar)
            cmd = [sz_bin, "x", "-y", "-bd", "-o%s" % cache_dir, inner_tar]
            logger.debug("Executing shell command `%s`" % " ".join(cmd))
            try:
                output = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
            except subprocess.CalledProcessError as e:
                logger.error("7zip failed: %s" % repr(e.output))
                raise Exception("Unable to extract inner Firefox archive")
            logger.debug("7zip succeeded at second pass: %s" % repr(output))
            os.remove(inner_tar)

    app = FirefoxApp(cache_dir)

    # Workaround until 7zip learns to maintain file attributes
    os.chmod(app.exe, stat.S_IREAD | stat.S_IEXEC | stat.S_IWUSR)

    return app



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import logging
import time
from worq.pool.thread import WorkerPool
from worq import get_broker, get_queue, TaskSpace

import xpcshell_worker as xw


logger = logging.getLogger(__name__)
ts = TaskSpace(__name__)
pool = None


def init(worq_url):
    global ts
    broker = get_broker(worq_url)
    broker.expose(ts)
    return broker


def start_pool(worq_url, num_workers=1, **kw):
    broker = init(worq_url)
    new_pool = WorkerPool(broker, workers=num_workers)
    new_pool.start(**kw)
    return new_pool


def stop():
    global logger, pool
    logger.debug("Stopping worker pool %s" % pool)
    if pool is not None:
        pool.stop()
        pool = None


class ScanResult(object):
    """Class to hold and evaluate scan responses."""

    def __init__(self, response):
        self.response = response
        self.success = self.evaluate_success(response)
        self.host = self.get_host()
        self.rank = self.get_rank()

    def get_host(self):
        return self.response.original_cmd["args"]["host"]

    def get_rank(self):
        return self.response.original_cmd["args"]["rank"]

    @staticmethod
    def evaluate_success(response):
        global logger

        # if .success is true, the result came through the requests
        # `load` handler with state == 4 (fully loaded).
        if response.success:
            return True

        # Else, check whether the error was due to a redirect error, with the first hop being OK.
        uri = response.result["info"]["original_uri"]
        status = response.result["info"]["status"]
        origin = response.result["origin"]
        if origin == "error_handler" and status == 0:  # NS_OK
            logger.debug("Ignored redirect by `%s`" % uri)
            return True

        # Else, the request had some sort of issue
        return False

    def as_dict(self):
        return {
            "response": self.response.as_dict(),
            "success": self.success,
            "host": self.host,
            "rank": self.rank
        }


@ts.task
def scan_urls(app, target_list, profile=None, prefs=None, get_certs=False, timeout=10):
    global logger

    logger.debug("scan_urls task called with %s" % repr(target_list))

    # Spawn a worker instance
    xpcw = xw.XPCShellWorker(app, profile=profile, prefs=prefs)
    xpcw.spawn()

    # Enqueue all host scans for this worker instance
    wakeup_cmd = xw.Command("wakeup")
    cmd_count = 0
    for rank, host in target_list:
        scan_cmd = xw.Command("scan", host=host, rank=rank, include_certificates=get_certs, timeout=timeout)
        xpcw.send(scan_cmd)
        if cmd_count % 10 == 0:
            xpcw.send(wakeup_cmd)
        cmd_count += 1

    # Fetch results from queue, until all results are in or until the last
    # scan must have into timeout. Note that ACKs come in strict sequence of
    # their respective commands.
    results = {}
    timeout_time = time.time() + timeout + 1
    while time.time() < timeout_time:
        for response in xpcw.receive():
            if response.result == "ACK":
                # Reset timeout when scan commands are ACKed.
                if response.original_cmd["mode"] == "scan":
                    timeout_time = time.time() + timeout + 1
                # Ignore other ACKs.
                continue
            # Else we know this is the result of a scan command.
            result = ScanResult(response)
            results[result.host] = result
        if len(results) >= len(target_list):
            break
        if xpcw.send(wakeup_cmd):
            time.sleep(0.1)
        else:
            break

    if len(results) < len(target_list):
        logger.warning("Worker task dropped results, yielded %d instead of %d" % (len(results), len(target_list)))

    # Wind down the worker
    xpcw.send(xw.Command("quit"))
    xpcw.terminate()

    logger.debug("Worker task finished, returning %d results" % len(results))

    return results


@ts.task
def collect(result_dicts):
    combined_results = {}
    for result in result_dicts:
        combined_results.update(result)
    return combined_results


def __as_chunks(flat_list, chunk_size):
    for i in range(0, len(flat_list), chunk_size):
        yield flat_list[i:i + chunk_size]


# CAVE: run_scans is not re-entrant due to use of global variables.
def run_scans(app, target_list, profile=None, prefs=None, num_workers=4, targets_per_worker=50, worq_url="memory://",
              get_certs=False, timeout=10, progress_callback=None):
    global logger, pool

    pool = start_pool(worq_url, timeout=1, num_workers=num_workers)

    chunks = __as_chunks(target_list, targets_per_worker)
    try:
        queue = get_queue(worq_url, target=__name__)

        # Enqueue tasks to be executed in parallel
        scan_results = [queue.scan_urls(app, targets, profile=profile, prefs=prefs,
                                        get_certs=get_certs, timeout=timeout)
                        for targets in chunks]
        result = queue.collect(scan_results)

        queue_len = len(queue)
        logged_len = 0  # Required to correct for "overlogging" due to chunking

        while True:
            finished = result.wait(timeout=10)
            current_queue_len = len(queue)
            chunks_done = queue_len - current_queue_len
            logger.debug("After queue wait: %d old - %d new = %d done" % (queue_len, current_queue_len, chunks_done))
            queue_len = current_queue_len
            # Check finished first to ensure that the final chunk is not logged,
            # because the final chunk might not have the full chunk size.
            if finished:
                break
            if progress_callback is not None and chunks_done > 0:
                # We must assume the maximum chunk size here to calculate the number of results
                progress_callback(chunks_done * targets_per_worker)
                logged_len += chunks_done * targets_per_worker

    except KeyboardInterrupt:
        logger.critical("Ctrl-C received. Winding down workers...")
        stop()
        logger.debug("Signaled workers to quit")
        raise KeyboardInterrupt

    finally:
        stop()

    # Log the results of the final chunk
    if progress_callback is not None:
        actual_len = len(result.value)
        logger.debug("Chunkwise logging reported on %d results, actually received %d" % (logged_len, actual_len))
        len_correction = actual_len - logged_len
        if len_correction != 0:
            logger.debug("Logging correction for %d results" % len_correction)
            progress_callback(len_correction)

    return result.value



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import logging
import threading
import time

logger = logging.getLogger(__name__)


class ProgressTracker(object):
    """Class that implements progress tracking"""

    def __init__(self, total, unit=None, percent=True, speed=True, eta=True, average=30*60.0):
        """
        Constructor for a ProcessLogger

        The class keeps track of completed and overhead calculations, and
        can estimate current progress speed and time of completion (ETA).

        :param total: int total calculations required for completion
        :param unit: str for unit used for speed
        :param percent: bool whether to show percent
        :param speed: bool whether or not to show speed
        :param eta: bool whether or not to show ETA
        :param average: float length of averaging window
        """
        self.total = total
        self.completed = 0
        self.overhead = 0
        self.unit = "" if unit is None else " %s" % unit
        self.show_percent = percent
        self.show_speed = speed
        self.show_eta = eta
        self.average_window = average
        self.log = []
        self.write_lock = threading.Lock()
        self.start_time = time.time()
        self.logger_thread = None

    def log_completed(self, completed):
        """
        Log a number of completed calculations

        :param completed: Number of completed items
        :return: None
        """
        self.write_lock.acquire()
        try:
            self.log.append((time.time(), completed, 0))
            self.completed += completed
        finally:
            self.write_lock.release()

    def log_overhead(self, overhead):
        """
        Log a number of overhead calculations
        :param overhead:
        :return:
        """
        self.write_lock.acquire()
        try:
            self.log.append((time.time(), 0, overhead))
            self.overhead += overhead
        finally:
            self.write_lock.release()

    def log_window(self, window):
        """
        Return current averaging window

        :param window: float time span for window
        :return: list of (time, int completed, int overhead)
        """
        earliest_time = time.time() - window
        latest_entry = len(self.log)
        earliest_entry = latest_entry
        for i in xrange(latest_entry - 1, -1, -1):
            if self.log[i][0] >= earliest_time:
                earliest_entry = i
            else:
                break
        return self.log[earliest_entry:latest_entry]

    @staticmethod
    def __window_parameters(log_window):
        """
        Calculate timespan and sums of completed and overhead for log window

        :param log_window: list
        :return: float time span, int completed, int overhead
        """
        # Skip first log entry in window, because it logs values from before target timeframe
        earliest, _, _ = log_window[0]
        latest, completed, overhead = reduce(lambda x, y: (y[0], x[1] + y[1], x[2] + y[2]), log_window[1:])
        span = latest - earliest
        return span, completed, overhead

    def __str__(self):
        """
        Return string representation of current progress

        :return: str
        """

        now = time.time()
        overall_time = now - self.start_time

        # Calculate overall progress and percentages
        net_total = self.total
        net_done = self.completed
        # net_todo = self.total - net_done
        net_percent = min(100.0 * net_done / net_total, 100.0)

        # Gross values take overhead into account
        gross_done = self.completed + self.overhead
        # Gross total is estimated by total overhead so far
        gross_total = net_total if net_done == 0 else net_total * gross_done / net_done
        gross_todo = gross_total - gross_done

        # Overhead is relative to net total
        overhead_percent = 0.0 if net_done == 0 else 100.0 * self.overhead / net_done

        # Get current averaging window
        log_window = self.log_window(self.average_window)

        # Bail out if there is not enough data in the window
        if len(log_window) < 2:
            s = u""
            if self.show_percent:
                s += u"%.0f%% " % net_percent
            s += u"%d/%d" % (min(net_done, net_total), net_total)
            s += u", %.0f%% overhead" % overhead_percent
            if self.show_speed:
                s += u", --%s/s net" % self.unit
                s += u", --%s/s gross" % self.unit
            if self.show_eta:
                s += u", ETA --"
            return s

        # Get values for current averaging window
        win_span, win_completed, win_overhead = self.__window_parameters(log_window)

        # Calculate overall and current net and gross speeds, and ETA
        # net_speed = float(net_done) / overall_time
        gross_speed = float(gross_done) / overall_time
        net_win_speed = float(win_completed) / win_span
        gross_win_speed = float(win_completed + win_overhead) / win_span
        # Speed might have been zero
        if gross_win_speed != 0.0:
            gross_eta = now + gross_todo / gross_win_speed
        else:
            gross_eta = None

        # Build the string according to config
        s = u""
        if self.show_percent:
            s += u"%.0f%% " % net_percent
        s += u"%d/%d" % (min(net_done, net_total), net_total)
        s += u", %.1f%% overhead" % overhead_percent
        if self.show_speed:
            match = [
                (0.001, "%s/ms" % self.unit),
                (1.0, "%s/s" % self.unit),
                (60.0, "%s/min" % self.unit),
                (60.0 * 60.0, "%s/h" % self.unit),
                (24.0 * 60.0 * 60.0, "%s/day" % self.unit)
            ]
            scale, unit = match[1]
            for scale, unit in match:
                if gross_speed * scale > 100:
                    break
            s += u", %.0f%s net" % (scale * net_win_speed, unit)
            s += u", %.0f%s gross" % (scale * gross_win_speed, unit)
        if self.show_eta:
            if gross_eta is not None:
                s += u", ETA %s" % time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime(gross_eta))
            else:
                s += u", ETA --"
        return s

    def start_reporting(self, interval, first_interval=None):
        """
        Spawn logger thread

        :param interval: float seconds between log update
        :param first_interval: optional float seconds before first log update
        :return: None
        """
        global logger

        if self.logger_thread is None:
            logger.debug("Starting progress logger thread")
            self.logger_thread = ProgressLogger(self, first_interval, interval)
            self.logger_thread.setName("Progress")
            self.logger_thread.daemon = True  # Thread dies with worker
            self.logger_thread.start()

    def stop_reporting(self):
        """
        Stop logger thread

        :return: None
        """
        global logger

        if self.logger_thread is not None:
            logger.debug("Terminating progress logger thread")
            try:
                self.logger_thread.quit()
            except AttributeError:
                pass
            finally:
                self.logger_thread = None


class ProgressLogger(threading.Thread):
    """Progress logger thread that logs progress updates"""

    def __init__(self, pr, first_interval, regular_interval):
        """
        Constructor

        :param pr: ProgressTracker instance to monitor
        :param first_interval: float seconds before first log update
        :param regular_interval: float seconds between log update
        """
        super(ProgressLogger, self).__init__()
        self.pr = pr
        self.first_interval = first_interval
        self.regular_interval = regular_interval
        self.updated_at = time.time()
        self.__quit = False

    def __update_time(self):
        """
        Iterator that returns time of next progress update.
        The first update may be different from the rest.

        :return: float iterator
        """
        if self.first_interval is not None:
            yield self.updated_at + self.first_interval
        while True:
            yield self.updated_at + self.regular_interval

    def run(self):
        """
        Start thread

        :return: None
        """
        global logger

        logger.debug("ProgressLogger thread starting")
        update_time = self.__update_time()
        next_update = update_time.next()
        while not self.__quit:
            time.sleep(1)
            now = time.time()
            if now >= next_update:
                logger.info(str(self.pr))
                self.updated_at = now
                next_update = update_time.next()
        logger.debug("ProgressLogger thread exiting")

    def quit(self):
        """
        Signal thread to terminate

        :return: None
        """
        self.__quit = True



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import dateutil.parser
from distutils import dir_util
import json
import logging
import os
import shutil

import cert


logger = logging.getLogger(__name__)
module_dir = os.path.split(__file__)[0]


def generate(mode, logs, output_dir):
    global logger

    logger.debug("Generating `%s` report for %d logs in `%s`" % (mode, len(logs), output_dir))

    if mode == "web":
        for log_name in sorted(logs.keys()):
            log = logs[log_name]
            meta = log.get_meta()
            if meta["mode"] != "regression":
                logger.warning("Skipping report generation for non-regression log `%s`" % log_name)
                continue
            if not log.has_finished():
                logger.warning("Skipping report generation for incomplete log `%s`" % log_name)
                continue
            if not log.is_compatible():
                logger.warning("Skipping report generation for incompatible log `%s`" % log_name)
                continue
            web_report(log, output_dir)
    else:
        logger.critical("Report generator mode `%s` not implemented" % mode)


def web_report(log, report_dir):
    global logger

    # Create report directory if necessary.
    if not os.path.exists(report_dir):
        logger.debug('Creating report directory %s' % report_dir)
        os.makedirs(report_dir)

    # Fetch log metadata
    meta = log.get_meta()
    run_start_time = dateutil.parser.parse(meta["run_start_time"])
    timestamp = run_start_time.strftime("%Y-%m-%d-%H-%M-%S")

    # Read the complete runs log to see if this log was already reported
    runs_log_file = os.path.join(report_dir, "runs", "runs.json")

    if os.path.exists(runs_log_file):
        with open(runs_log_file) as f:
            runs_log = json.load(f)
            for line in runs_log[0]["data"]:
                logger.debug("Line read from runs.json: `%s`" % line)
    else:
        # File does not exist, create an empty log
        runs_log = json.loads('[{"data":[]}]')

    if timestamp in json.dumps(runs_log):
        logger.warning("Skipping log `%s` which was already reported before" % log.handle)
        return

    # Write log file
    run_dir = os.path.join(report_dir, "runs", timestamp)
    logger.info("Writing HTML report to `%s`" % run_dir)

    uri_data = []
    for line in log:
        if meta["args"]["filter"] == 1:
            # Filter out stray timeout errors
            connection_speed = line["response"]["response_time"]-line["response"]["command_time"]
            timeout = line["response"]["original_cmd"]["args"]["timeout"] * 1000
            error_message = line["response"]["result"]["info"]["short_error_message"]
            if error_message == "NS_BINDING_ABORTED" and connection_speed > timeout:
                continue
        uri_data.append(line)

    log_data = [{"meta": log.get_meta(), "data": uri_data}]

    # Install static template files in report directory
    template_dir = os.path.join(module_dir, "template")
    dir_util.copy_tree(os.path.join(template_dir, "js"),
                       os.path.join(report_dir, "js"))
    dir_util.copy_tree(os.path.join(template_dir, "css"),
                       os.path.join(report_dir, "css"))
    dir_util.copy_tree(os.path.join(template_dir, "img"),
                       os.path.join(report_dir, "img"))
    shutil.copyfile(os.path.join(template_dir, "index.htm"),
                    os.path.join(report_dir, "index.htm"))

    # Create per-run directory for report output
    if not os.path.isdir(run_dir):
        os.makedirs(run_dir)

    # Copy profiles
    if "profiles" in meta:
        for profile in meta["profiles"]:
            log_zip = log.part(profile["log_part"])
            run_dir_zip = os.path.join(run_dir, profile["log_part"])
            logger.debug("Copying `%s` profile archive from `%s` to `%s`" % (profile["name"], log_zip, run_dir_zip))
            shutil.copyfile(log_zip, run_dir_zip)

    cert_dir = os.path.join(run_dir, "certs")
    __extract_certificates(log, cert_dir)

    shutil.copyfile(os.path.join(template_dir, "report_template.htm"),
                    os.path.join(run_dir, "index.htm"))

    # Write the final log file
    with open(os.path.join(run_dir, "log.json"), "w") as log_file:
        log_file.write(json.dumps(log_data, indent=4, sort_keys=True))

    # Append to runs log
    new_run_log = {
            "run": timestamp,
            "branch": meta["test_metadata"]["branch"].capitalize(),
            "errors": len(log),
            "description": "Fx%s %s vs Fx%s %s" % (meta["test_metadata"]["app_version"],
                                                   meta["test_metadata"]["branch"],
                                                   meta["base_metadata"]["app_version"],
                                                   meta["base_metadata"]["branch"])
        }
    runs_log[0]["data"].append(new_run_log)
    logger.debug("Writing back runs log to `%s`" % runs_log_file)
    with open(runs_log_file, "w") as f:
        f.write(json.dumps(runs_log, indent=4, sort_keys=True))


def __extract_certificates(log, cert_dir):
    global logger

    if not os.path.exists(cert_dir):
        os.makedirs(cert_dir)

    for log_line in log:
        result = {
            "host": log_line["host"],
            "rank": log_line["rank"],
            "response": log_line["response"]
        }
        cert_file = os.path.join(cert_dir, "%s.der" % result["host"])
        if "certificate_chain" in result["response"]["result"]["info"]:
            server_cert_string = "".join(map(chr, result["response"]["result"]["info"]["certificate_chain"][0]))
            logger.debug("Writing certificate data for `%s` to `%s`" % (result["host"], cert_file))
            with open(cert_file, "w") as f:
                f.write(server_cert_string)
        else:
            logger.debug("No certificate data available for `%s`" % result["host"])


NSErrorMap = {
    # For network error messages that are not obtainable otherwise
    # https://developer.mozilla.org/en-US/docs/Mozilla/Errors
    0X00000000: "NS_OK",
    0X80004004: "NS_ERROR_ABORT",
    0X8000FFFF: "UNEXPECTED_ERROR",
    0X804B0002: "NS_BINDING_ABORTED",
    0X804B000A: "ERROR_MALFORMED_URI",
    0X804B000D: "CONNECTION_REFUSED_ERROR",
    0X804B0014: "NET_RESET_ERROR",
    0X804B001E: "DOMAIN_NOT_FOUND_ERROR",
}


def decode_ns_status(scan_result):
    status = scan_result["response"]["result"]["info"]["status"]
    try:
        return NSErrorMap[status]
    except KeyError:
        return "UNKNOWN_STATUS"


def decode_error_type(scan_result):
    status = scan_result["response"]["result"]["info"]["status"]
    if status & 0xff0000 == 0x5a0000:  # security module
        error_class = scan_result["response"]["result"]["info"]["error_class"]
        if error_class == 2:  # nsINSSErrorsService::ERROR_CLASS_BAD_CERT
            return "certificate"
        else:
            return "protocol"
    else:
        return "network"


def decode_raw_error(scan_result):
    if "raw_error" in scan_result["response"]["result"]["info"]:
        raw_error = scan_result["response"]["result"]["info"]["raw_error"]
        if "Error code:" in raw_error:
            return raw_error.split("Error code:")[1].split(">")[1].split("<")[0]
    return decode_ns_status(scan_result)


def collect_error_info(scan_result):
    error_info = {
        "message": decode_raw_error(scan_result),
        "code": "%s" % hex(scan_result["response"]["result"]["info"]["status"]),
        "type": decode_error_type(scan_result)
    }
    return error_info


def collect_site_info(scan_result):
    site_info = {
        "timestamp": scan_result["response"]["response_time"],
        "connectionSpeed": scan_result["response"]["response_time"] - scan_result["response"]["command_time"],
        "uri": scan_result["host"],
        "rank": scan_result["rank"]
    }
    return site_info


def collect_certificate_info(scan_result):

    result = scan_result["response"]["result"]

    if not result["info"]["ssl_status_status"]:
        return {}

    status = result["info"]["ssl_status"]

    server_cert = status["serverCert"]
    parsed_server_cert = cert.Cert(result["info"]["certificate_chain"][0])

    root_cert = server_cert
    chain_length = 1
    while root_cert["issuer"] is not None:
        root_cert = root_cert["issuer"]
        chain_length += 1

    cert_info = {
        "nickname": server_cert["nickname"] if "nickname" in server_cert else "(no nickname)",
        "emailAddress": server_cert["emailAddress"],
        "subjectName": server_cert["subjectName"],
        "commonName": server_cert["commonName"],
        "organization": server_cert["organization"],
        "organizationalUnit": server_cert["organizationalUnit"],
        "issuerCommonName": server_cert["issuerCommonName"],
        "issuerOrganization": server_cert["issuerOrganization"],
        "sha1Fingerprint": server_cert["sha1Fingerprint"],
        "sha256Fingerprint": server_cert["sha256Fingerprint"],
        "chainLength": chain_length,
        "certifiedUsages": result["info"]["certified_usages"],
        "validityNotBefore": server_cert["validity"]["notBeforeGMT"],
        "validityNotAfter": server_cert["validity"]["notAfterGMT"],
        "isEV": str(status["isExtendedValidation"]),
        "subjectAltName": parsed_server_cert.subject_alt_name(),
        "signatureAlgorithm": parsed_server_cert.signature_hash_algorithm(),
        "keyUsage": server_cert["keyUsages"],
        "extKeyUsage": parsed_server_cert.ext_key_usage(),
        "rootCertificateSubjectName": root_cert["subjectName"],
        "rootCertificateOrganization": root_cert["organization"],
        "rootCertificateOrganizationalUnit": root_cert["organizationalUnit"],
        "rootCertificateSHA1Fingerprint": root_cert["sha1Fingerprint"],
    }

    return cert_info


def collect_scan_info(scan_result):
    return {
        "site_info": collect_site_info(scan_result),
        "error": collect_error_info(scan_result),
        "cert_info": collect_certificate_info(scan_result)
    }


def add_performance_info(log_data, scan_result):
    log_data["site_info"]["connectionSpeedChange"] = scan_result["response"]["connection_speed_change"]
    log_data["site_info"]["connectionSpeedSamples"] = scan_result["response"]["connection_speed_samples"]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

from setuptools import setup, find_packages

PACKAGE_NAME = 'tlscanary'
PACKAGE_VERSION = '3.2.0'

INSTALL_REQUIRES = [
    'coloredlogs',
    'cryptography',
    'hashfs',
    'python-dateutil',
    'worq'
]

TESTS_REQUIRE = [
    'coverage',
    'nose',
    'mock'
]

DEV_REQUIRES = [
    'coverage',
    'nose',
    'mock',
    'pep8'
]

setup(
    name=PACKAGE_NAME,
    version=PACKAGE_VERSION,
    description='TLS/SSL Test Suite for Mozilla Firefox',
    classifiers=[
        'Environment :: Console',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows :: Windows 10',
        'Operating System :: Microsoft :: Windows :: Windows 7',
        'Operating System :: Microsoft :: Windows :: Windows 8',
        'Operating System :: Microsoft :: Windows :: Windows 8.1',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Quality Assurance',
        'Topic :: Software Development :: Testing'
    ],
    keywords=['mozilla', 'firefox', 'tls', 'regression-testing', 'testing'],
    author='Christiane Ruetten',
    author_email='cr@mozilla.com',
    url='https://github.com/mozilla/tls-canary',
    download_url='https://github.com/mozilla/tls-canary/archive/latest.tar.gz',
    license='MPL2',
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,  # See MANIFEST.in
    zip_safe=False,
    use_2to3=False,
    install_requires=INSTALL_REQUIRES,
    tests_require=TESTS_REQUIRE,
    extras_require={'dev': DEV_REQUIRES},  # For `pip install -e .[dev]`
    test_suite='nose.collector',
    entry_points={
        'console_scripts': [
            'tlscanary = tlscanary.main:main'
        ]
    }
)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import basemode
import performance
import regression
import log
import scan
import sourceupdate

__all__ = ["log", "performance", "regression", "scan", "sourceupdate"]


def __subclasses_of(cls):
    sub_classes = cls.__subclasses__()
    sub_sub_classes = []
    for sub_cls in sub_classes:
        sub_sub_classes += __subclasses_of(sub_cls)
    return sub_classes + sub_sub_classes


# Keep a record of all BaseMode subclasses
all_modes = dict([(mode.name, mode) for mode in __subclasses_of(basemode.BaseMode)])
all_mode_names = sorted(all_modes.keys())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import ConfigParser
import glob
import os
import struct


class FirefoxApp(object):
    """Class holding information about an extracted Firefox application directory"""

    __locations = {
        "osx": {
            "base": os.path.join("*", "*.app"),
            "exe": os.path.join("Contents", "MacOS", "firefox"),
            "browser": os.path.join("Contents", "Resources", "browser"),
            "gredir": os.path.join("Contents", "Resources"),
            "ini": os.path.join("Contents", "Resources", "application.ini"),
        },
        "linux": {
            "base": "firefox",
            "exe": "firefox",
            "browser": "browser",
            "gredir": "",
            "ini": "application.ini",
        },
        "win": {
            "base": "core",
            "exe": "firefox.exe",
            "browser": "browser",
            "gredir": "",  # FIXME: this may be wrong
            "ini": "application.ini",
        },
    }

    def __init__(self, directory):

        # Assuming that directory points to a directory
        # where a stock Firefox archive was extracted.
        self.platform = None
        self.app_dir = None
        for platform in self.__locations:
            base = self.__locations[platform]["base"]
            matches = glob.glob(os.path.join(directory, base))
            if len(matches) == 0:
                continue
            elif len(matches) >= 1:
                if os.path.isdir(matches[0]):
                    self.platform = platform
                    self.app_dir = matches[0]
                    break
            raise Exception("Unsupported application package format (missing or ambiguous base folder)")

        if self.platform is None:
            raise Exception("Unsupported application package platform")

        # Fill in the rest of the package locations
        self.exe = os.path.join(self.app_dir, self.__locations[self.platform]["exe"])
        self.browser = os.path.join(self.app_dir, self.__locations[self.platform]["browser"])
        self.gredir = os.path.join(self.app_dir, self.__locations[self.platform]["gredir"])
        self.app_ini = os.path.join(self.app_dir, self.__locations[self.platform]["ini"])

        # Sanity checks
        if not os.path.isfile(self.exe) or not os.path.isdir(self.browser):
            raise Exception("Unsupported application package format (missing files)")

        # For `linux`: byte 4 in ELF header is 01/02 for 32/64 bit
        if self.platform == "linux":
            with open(self.exe) as f:
                head = f.read(5)
            if head[4] == '\x01':
                self.platform = "linux32"
            elif head[4] == '\x02':
                self.platform = "linux"
            else:
                raise Exception("Unsupported ELF binary (%s)" % ord(head[4]))
        # Windows PE header offset is at 0x3c. Bytes 4 and 5 there tell 32 from 64 bit
        elif self.platform == "win":
            with open(self.exe) as f:
                try:
                    f.seek(0x3c)
                    pe_header_offset = struct.unpack("<I", f.read(4))[0]
                    f.seek(pe_header_offset)
                    pe_sig, machine_type = struct.unpack("<4sH", f.read(6))
                except struct.error:
                    # Definitely invalid
                    pe_sig, machine_type = (b'', 0)
                if (pe_sig, machine_type) == (b"PE\x00\x00", 0x14c):
                    self.platform = "win32"
                elif (pe_sig, machine_type) == (b"PE\x00\x00", 0x8664):
                    self.platform = "win"
                else:
                    raise Exception("Unsupported PE binary format")

        # Determine Firefox version
        ini_parser = ConfigParser.SafeConfigParser()
        ini_parser.read(self.app_ini)
        self.application_ini = dict(ini_parser.items("App"))
        self.version = ini_parser.get("App", "Version")
        # For versions that have no `CodeName` specified, extract it from the repo name.
        try:
            self.release = ini_parser.get("App", "CodeName")
        except ConfigParser.NoOptionError:
            self.release = ini_parser.get("App", "sourcerepository").split("-")[-1]

        # Field for optional package origin metadata (must be provided externally)
        self.package_origin = None
